import streamlit as st


from autolamella import config as cfg
import os

import plotly.io as pio
import pandas as pd

import plotly.express as px

st.set_page_config(layout="wide")

pio.templates.default = "plotly_white"

conn = st.experimental_connection("fibsem.db", type="sql", url = "sqlite:///fibsem.db")

# select project
df = conn.query("SELECT * from projects")

PROJECT_NAMES = df["name"].values.tolist()
PROJECT_NAME = st.sidebar.selectbox("Select project", PROJECT_NAMES)

PROJECT_ID = df[df["name"] == PROJECT_NAME]["id"].values[0]

# select experiment from project
df = conn.query(f"SELECT * FROM experiments WHERE project_id = {PROJECT_ID}")

EXPERIMENT_NAMES = df["name"].values.tolist()
EXPERIMENT_NAME = st.sidebar. selectbox("Select experiment", EXPERIMENT_NAMES)
EXPERIMENT_ID = df[df["name"] == EXPERIMENT_NAME]["id"].values[0]

# show data from experiment
# df = conn.query(f"SELECT * FROM experiments WHERE project_id = {PROJECT_ID} AND name = '{EXPERIMENT_NAME}'")


# show data from experiment

st.header(f"Experiment: {EXPERIMENT_NAME}")


# HISTORY
df_history = conn.query(f"SELECT e.name, e.date, h.petname, h.start, h.end, h.duration, h.stage FROM history h JOIN experiments e ON e.id = h.experiment_id WHERE h.experiment_id = {EXPERIMENT_ID} ")


# Duration
cols = st.columns(2)
fig_duration = px.bar(df_history, x="petname", y="duration", color="stage", barmode="group", hover_data=df_history.columns, title="Lamella Duration by Stage")
cols[0].plotly_chart(fig_duration, use_container_width=True)

# Duration
fig_duration = px.bar(df_history, x="stage", y="duration", color="petname", barmode="group", hover_data=df_history.columns, title="Stage Duration by Lamella")
cols[1].plotly_chart(fig_duration, use_container_width=True)

# STEPS
df_steps = conn.query(f"SELECT * FROM steps WHERE experiment_id = {EXPERIMENT_ID}")

# convert timestamp to datetime, aus timezone 
df_steps.timestamp = pd.to_datetime(df_steps.timestamp)

# convert timestamp to australian timezone
df_steps.timestamp = df_steps.timestamp.dt.tz_convert("Australia/Sydney")

fig_timeline = px.scatter(df_steps, x="step_n", y="timestamp", color="stage", symbol="petname",
    title="AutoLamella Timeline", 
    hover_name="stage", hover_data=df_steps.columns)
    # size = "duration", size_max=20)
st.plotly_chart(fig_timeline, use_container_width=True)



# DETECTIONS
df_det = conn.query(f"SELECT * FROM detections WHERE experiment_id = {EXPERIMENT_ID}")


df_group = df_det.groupby(["feature", "is_correct"]).count().reset_index() 
df_group = df_group.pivot(index="feature", columns="is_correct", values="petname")

# if no false, add false column
if "False" not in df_group.columns:
    df_group["False"] = 0
if "True" not in df_group.columns:
    df_group["True"] = 0

# fill missing values with zero
df_group.fillna(0, inplace=True)

df_group["total"] = df_group["True"] + df_group["False"]
df_group["percent_correct"] = df_group["True"] / df_group["total"]
df_group["percent_correct"] = df_group["percent_correct"]#.round(2)
# df_group = df_group.sort_values(by="percent_correct", ascending=False)
df_group.reset_index(inplace=True)

cols = st.columns(2)
# plot
fig_acc = px.bar(df_group, x="feature", y="percent_correct", color="feature", title="ML Accuracy", hover_data=df_group.columns)
cols[0].plotly_chart(fig_acc, use_container_width=True)

# precision
fig_det = px.scatter(df_det, x="dpx_x", y="dpx_y", color="stage", symbol="feature",  hover_data=df_det.columns, title="ML Error Size")
cols[1].plotly_chart(fig_det, use_container_width=True)

# INTERACTIONS

df = conn.query(f"""SELECT e.name, e.date, i.timestamp, i.petname, i.stage, i.step, i.type, i.subtype, i.dm_x, i.dm_y, i.beam_type FROM interactions i JOIN experiments e ON e.id = i.experiment_id""")


# plot
fig_interactions = px.scatter(df, x="dm_x", y="dm_y", color="type", symbol="subtype", hover_data=df.columns, title="Interactions")
st.plotly_chart(fig_interactions, use_container_width=True)



# SPLIT INTO RUNS / AGGREGATE DATA



st.markdown("---")
st.header("Comparative Data")

tab_ml, tab_duration = st.tabs(["Machine Learning", "Duration"])

with tab_ml:
    st.subheader("Machine Learning")

    df = conn.query(f"SELECT e.name, e.date, d.timestamp, d.petname, d.stage, d.step, d.feature, d.is_correct, d.dpx_x, d.dpx_y FROM detections d JOIN experiments e ON e.id = d.experiment_id")
    

    # drop if Autoliftout in e.name
    df = df[~df["name"].str.contains("AUTOLIFTOUT")]

    UNIQUE_FEATURES_IN_EXPERIMENT = df[df["name"] == EXPERIMENT_NAME]["feature"].unique()

    # filter df
    df = df[df["feature"].isin(UNIQUE_FEATURES_IN_EXPERIMENT)]

    df_group = df.groupby(["name", "date", "feature", "is_correct"]).count().reset_index()

    df_group = df_group.pivot(index=["name", "date", "feature"], columns="is_correct", values="petname")

    # if no false, add false column
    if "False" not in df_group.columns:
        df_group["False"] = 0
    if "True" not in df_group.columns:
        df_group["True"] = 0

    # fill missing values with zero
    df_group.fillna(0, inplace=True)

    df_group["total"] = df_group["True"] + df_group["False"]
    df_group["percent_correct"] = df_group["True"] / df_group["total"]
    df_group["percent_correct"] = df_group["percent_correct"]#.round(2)
    df_group = df_group.sort_values(by="date", ascending=False)

    df_group.reset_index(inplace=True)

    # plot
    # sort by date
    df_group.sort_values(by="date", ascending=True, inplace=True)
    fig_acc = px.bar(df_group, x="name", y="percent_correct", color="feature", barmode="group", title="ML Accuracy", hover_data=df_group.columns)
    st.plotly_chart(fig_acc, use_container_width=True)

    # plot line chart
    fig_acc = px.line(df_group, x="name", y="percent_correct", color="feature", title="ML Accuracy", hover_data=df_group.columns)

    # set y limits at 0 - 1
    fig_acc.update_yaxes(range=[0, 1])
    st.plotly_chart(fig_acc, use_container_width=True)

    # group by name, petname, feature, is_correct
    df_group = df.groupby(["name", "date", "petname", "stage", "feature", "is_correct"]).count().reset_index()
    df_group = df_group.pivot(index=["name", "date", "petname", "stage", "feature"], columns="is_correct", values="timestamp")

    # if no false, add false column
    if "False" not in df_group.columns:
        df_group["False"] = 0
    if "True" not in df_group.columns:
        df_group["True"] = 0

    # fill missing values with zero
    df_group.fillna(0, inplace=True)

    df_group["total"] = df_group["True"] + df_group["False"]
    df_group["percent_correct"] = df_group["True"] / df_group["total"]
    df_group["percent_correct"] = df_group["percent_correct"]#.round(2)
    df_group = df_group.sort_values(by="date", ascending=False)

    df_group.reset_index(inplace=True)

    st.dataframe(df_group, use_container_width=True)

    # sort by date, petname
    df_group.sort_values(by=["date", "petname"], ascending=True, inplace=True)

    for STAGE_NAME in df_group["stage"].unique():
        df_group_stage_filt = df_group[df_group["stage"] == STAGE_NAME]
        fig_ml = px.bar(df_group_stage_filt, x="petname", y="percent_correct", color="feature", 
                        barmode="group", facet_row="name", title=f"ML Accuracy Per Lamella Per Experiment (Stage: {STAGE_NAME})", hover_data=df_group.columns, 
                        height=max(100*len(df_group_stage_filt["name"].unique()), 300))
        
        # set y limits 0 -1
        fig_ml.update_yaxes(range=[0, 1])
        st.plotly_chart(fig_ml, use_container_width=True)

    # select all detections
    df = conn.query(f"SELECT e.name, e.date, d.timestamp, d.petname, d.stage, d.step, d.feature, d.is_correct, d.dpx_x, d.dpx_y FROM detections d JOIN experiments e ON e.id = d.experiment_id")
    df = df[~df["name"].str.contains("AUTOLIFTOUT")]

    # filter df
    df = df[df["feature"].isin(UNIQUE_FEATURES_IN_EXPERIMENT)]

    # calculate average accuracy
    df_group = df.groupby(["name", "date", "is_correct"]).count().reset_index()
    df_group = df_group.pivot(index=["name", "date"], columns="is_correct", values="petname")


    # if no false, add false column
    if "False" not in df_group.columns:
        df_group["False"] = 0
    if "True" not in df_group.columns:
        df_group["True"] = 0

    # fill missing values with zero
    df_group.fillna(0, inplace=True)

    df_group["total"] = df_group["True"] + df_group["False"]
    df_group["percent_correct"] = df_group["True"] / df_group["total"]
    df_group["percent_correct"] = df_group["percent_correct"]#.round(2)
    df_group = df_group.sort_values(by="date", ascending=False)

    df_group.reset_index(inplace=True)

    N_TRUE = df_group["True"].sum()
    N_FALSE = df_group["False"].sum()
    N_TOTAL = N_TRUE + N_FALSE
    ACCURACY = N_TRUE / N_TOTAL

    SELECTED_ACCURACY = df_group[df_group["name"] == EXPERIMENT_NAME]["percent_correct"].values[0]

    # UNIQUE_FEATURES_IN_EXPERIMENT = df[df["name"] == EXPERIMENT_NAME]["feature"].unique()

    cols = st.columns(len(UNIQUE_FEATURES_IN_EXPERIMENT)+1)
    cols[0].metric("ML Accuracy", f"{SELECTED_ACCURACY*100:.2f}%", delta=f"{(SELECTED_ACCURACY - ACCURACY)*100:.2f}%")

    for i, FEATURE_NAME in enumerate(UNIQUE_FEATURES_IN_EXPERIMENT, 1):
        df_det_filt = df[df["feature"] == FEATURE_NAME]

        # calculate average accuracy
        df_group = df_det_filt.groupby(["name", "date", "is_correct"]).count().reset_index()
        df_group = df_group.pivot(index=["name", "date"], columns="is_correct", values="petname")


        # if no false, add false column
        if "False" not in df_group.columns:
            df_group["False"] = 0
        if "True" not in df_group.columns:
            df_group["True"] = 0

        # fill missing values with zero
        df_group.fillna(0, inplace=True)

        df_group["total"] = df_group["True"] + df_group["False"]
        df_group["percent_correct"] = df_group["True"] / df_group["total"]
        df_group["percent_correct"] = df_group["percent_correct"]#.round(2)
        df_group = df_group.sort_values(by="date", ascending=False)

        df_group.reset_index(inplace=True)

        N_TRUE = df_group["True"].sum()
        N_FALSE = df_group["False"].sum()
        N_TOTAL = N_TRUE + N_FALSE
        ACCURACY = N_TRUE / N_TOTAL

        SELECTED_ACCURACY = df_group[df_group["name"] == EXPERIMENT_NAME]["percent_correct"].values[0]

        cols[i].metric(f"{FEATURE_NAME}", f"{SELECTED_ACCURACY*100:.2f}%", delta=f"{(SELECTED_ACCURACY - ACCURACY)*100:.2f}%")


# st.markdown("---")
with tab_duration:
    # history
    st.subheader("Stage Duration")
    df_history = conn.query(f"SELECT e.name, e.date, e.id, h.petname, h.start, h.end, h.duration, h.stage FROM history h JOIN experiments e ON e.id = h.experiment_id")

    UNIQUE_STAGES_IN_EXPERIMENT =  df_history[df_history["name"]==EXPERIMENT_NAME]["stage"].unique()

    SETUP_STAGES = ["SetupTrench", "ReadyTrench", "SetupLamella", "Finished", "Setup"]

    # filter out setup stages from unique stages
    UNIQUE_STAGES_IN_EXPERIMENT = [x for x in UNIQUE_STAGES_IN_EXPERIMENT if x not in SETUP_STAGES]

    # filter out other stages
    df_history = df_history[df_history["stage"].isin(UNIQUE_STAGES_IN_EXPERIMENT)]
    # df_history = df_history[~df_history["stage"].isin(SETUP_STAGES)] # FILTER OUT SETUP

    # calculate average duration # fill missing values with zero
    df_group = df_history.groupby(["name", "date", "stage"]).mean().reset_index()
    df_group.fillna(0, inplace=True)
    df_group.reset_index(inplace=True)

    # drop stages with setup, ready
    df_group = df_group[~df_group["stage"].isin(["SetupTrench", "ReadyTrench", "SetupLamella", "Finished", "Setup"])]
    df_group = df_group.sort_values(by="date", ascending=True)

    df_group["duration_mins"] = df_group["duration"] /60
    df_group["duration_hrs"] = df_group["duration"] /60/60

    # plot
    fig_duration = px.bar(df_group, x="name", y="duration_mins", color="stage", barmode="group", title="Duration", hover_data=df_group.columns)
    st.plotly_chart(fig_duration, use_container_width=True)

    fig_duration = px.line(df_group, x="name", y="duration_mins", color="stage", title="Stage Duration", hover_data=df_group.columns)

    # set y limits at 0 - 1
    # fig_acc.update_yaxes(range=[0, 1])
    st.plotly_chart(fig_duration, use_container_width=True)


    # calculate average duration per lamella
    df_history_filter = df_history[~df_history["stage"].isin(["SetupTrench", "ReadyTrench", "SetupLamella", "Finished", "Setup"])]



    st.subheader("Stage Duration")


    # OVERALL DURATION
    df_group = df_history_filter.groupby(["name", "date"]).mean().reset_index()
    df_group = df_group.sort_values(by="duration", ascending=False)

    df_group["duration_hrs"] = df_group["duration"] /60/60  
    df_group["duration_mins"] = df_group["duration"] /60
    # drop id column
    df_group.drop(columns="id", inplace=True)

    SELECTED_DURATION = df_group[df_group["name"] == EXPERIMENT_NAME]["duration"].values[0]
    AVERAGE_DURATION = df_group['duration'].mean()
    DURATION_DELTA = (SELECTED_DURATION - AVERAGE_DURATION) / AVERAGE_DURATION * 100


    cols = st.columns(len(UNIQUE_STAGES_IN_EXPERIMENT)+1)
    cols[0].metric(f"Average", f"{SELECTED_DURATION/60:.2f} mins", 
        delta=f"{DURATION_DELTA:.2f}%", delta_color="inverse")

    st.dataframe(df_group, use_container_width=True)

    # PER STAGE DURATION
    for i, STAGE_NAME in enumerate(UNIQUE_STAGES_IN_EXPERIMENT, 1):
        df_history_filter2 = df_history_filter[df_history_filter["stage"] == STAGE_NAME]

        df_group = df_history_filter2.groupby(["name", "date"]).mean().reset_index()
        df_group = df_group.sort_values(by="duration", ascending=False)


        # show metric
        try:
            SELECTED_DURATION = df_group[df_group["name"] == EXPERIMENT_NAME]["duration"].values[0]
            AVERAGE_DURATION = df_group['duration'].mean()
            DURATION_DELTA = (SELECTED_DURATION - AVERAGE_DURATION) / AVERAGE_DURATION * 100

            cols[i].metric(f"{STAGE_NAME}", f"{SELECTED_DURATION/60:.2f}min", 
                delta=f"{DURATION_DELTA:.2f}%", delta_color="inverse")
        except Exception as e:
            st.write(e)
            cols[i].empty()


import logging
import os
import sys
from pprint import pprint

import matplotlib.patches as mpatches
from fibsem import conversions, utils
from fibsem.structures import BeamType, MicroscopeSettings, Point
from fibsem.ui import utils as fibsem_ui_utils

from fibsem.detection import utils as det_utils
from fibsem.detection.utils import Point
from fibsem.detection.detection import DetectedFeatures
                                     

from fibsem.ui.qtdesigner_files import detection_dialog as detection_gui
from PyQt5 import QtCore, QtWidgets

# TODO: convert to napari
class FibsemDetectionUI(detection_gui.Ui_Dialog, QtWidgets.QDialog):
    def __init__(
        self,
        microscope,
        settings: MicroscopeSettings,
        detected_features: DetectedFeatures,
    ):
        super(FibsemDetectionUI, self).__init__()
        self.setupUi(self)
        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)

        # microscope settings
        self.microscope = microscope
        self.settings = settings
        self.log_path = os.path.dirname(settings.image.path)

        # detection data
        self.detected_features = detected_features
        self.selected_feature_idx = 0
        self.selected_feature = self.detected_features.features[
                self.selected_feature_idx
            ].type
        self.logged_detection_types = []

        # images
        self.image = self.detected_features.image
        self._USER_CORRECTED = False

        # pattern drawing
        self.wp = fibsem_ui_utils._WidgetPlot(self, display_image=self.image)
        self.label_image.setLayout(QtWidgets.QVBoxLayout())
        self.label_image.layout().addWidget(self.wp)
        self.wp.canvas.mpl_connect("button_press_event", self.on_click)

        self.setup_connections()
        self.update_display()

        AUTO_CONTINUE = False # TODO:
        if AUTO_CONTINUE:
            self.continue_button_pressed() # automatically continue

    def setup_connections(self):

        self.comboBox_detection_type.clear()
        self.comboBox_detection_type.addItems(
            [feature.type.name for feature in self.detected_features.features]
        )
        self.comboBox_detection_type.setCurrentText(
            self.detected_features.features[0].type.name
        )
        self.comboBox_detection_type.currentTextChanged.connect(
            self.update_detection_type
        )

        self.pushButton_continue.clicked.connect(self.continue_button_pressed)

    def update_detection_type(self):

        self.selected_feature_idx = self.comboBox_detection_type.currentIndex()
        self.selected_feature = self.detected_features.features[
            self.selected_feature_idx
        ].type
        logging.info(
            f"Changed to {self.selected_feature_idx}, {self.selected_feature.name}"
        )

    def on_click(self, event):
        """Redraw the patterns and update the display on user click"""
        if event.button == 1 and event.inaxes is not None:
            self.xclick = event.xdata
            self.yclick = event.ydata

            self.point_m = conversions.image_to_microscope_image_coordinates(Point(self.xclick, self.yclick), self.image, self.detected_features.pixelsize)

            logging.debug(
                f"""DectectedFeature {self.selected_feature.name} | IMAGE COORDINATE | {int(self.xclick)}, {int(self.yclick)} | REAL COORDINATE | {self.point_m.x:.2e}, {self.point_m.y:.2e}"""
                
            )
            logging.debug(f"DectectedFeature {self.selected_feature.name} | COORD {int(self.xclick)}, {int(self.yclick)} | SHIFT | {self.point_m.x:.2e}, {self.point_m.y:.2e} | {self.settings.image.beam_type}")

            # update detection data
            self.detected_features.features[
                self.selected_feature_idx
            ].px = Point(self.xclick, self.yclick)

            # logging statistics
            logging.debug(
                f"Feature | {self.selected_feature} | {False}"
            )

            self.logged_detection_types.append(self.selected_feature)

            # flag that the user corrected a detection.
            self._USER_CORRECTED = True

            self.update_display()

    def update_display(self):
        """Update the window display. Redraw the crosshair"""


        # point position, image coordinates
        point_1 = self.detected_features.features[0].px
        point_2 = self.detected_features.features[1].px

        # colours
        c1 = det_utils.DETECTION_TYPE_COLOURS[
            self.detected_features.features[0].type
        ]
        c2 = det_utils.DETECTION_TYPE_COLOURS[
            self.detected_features.features[1].type
        ]

        # redraw all crosshairs
        self.wp.canvas.ax11.patches.clear()

        # draw cross hairs
        fibsem_ui_utils.draw_crosshair(
            self.image, self.wp.canvas, x=point_1.x, y=point_1.y, colour=c1,
        )
        fibsem_ui_utils.draw_crosshair(
            self.image, self.wp.canvas, x=point_2.x, y=point_2.y, colour=c2,
        )

        # draw arrow
        fibsem_ui_utils.draw_arrow(point_1, point_2, self.wp.canvas)

        # legend
        patch_one = mpatches.Patch(
            color=c1, label=self.detected_features.features[0].type.name
        )
        patch_two = mpatches.Patch(
            color=c2, label=self.detected_features.features[1].type.name
        )
        self.wp.canvas.ax11.legend(handles=[patch_one, patch_two])

        # calculate movement distance
        point_diff_px  = conversions.distance_between_points(point_1, point_2)
        point_diff_m = conversions.convert_point_from_pixel_to_metres(point_diff_px, self.detected_features.pixelsize)
        self.detected_features.distance = point_diff_m  # move from 1 to 2 (reverse direction)

        # update labels
        self.label_movement_header.setText(f"Movement")
        self.label_movement_header.setStyleSheet("font-weight:bold")
        self.label_movement.setText(
            f"""Moving {self.detected_features.features[0].type.name} to {self.detected_features.features[1].type.name}
         \nx distance: {self.detected_features.distance.x*1e6:.2f}um 
         \ny distance: {self.detected_features.distance.y*1e6:.2f}um"""
        )

        self.wp.canvas.draw()
    
    def continue_button_pressed(self):

        self.close() #exit

    def closeEvent(self, event):
        """Override the close event to save the data"""
        # log active learning data...
        if self._USER_CORRECTED:
            path = os.path.join(self.log_path, "label")
            det_utils.write_data_to_disk(path, self.detected_features)

        # log correct detection types
        for feature in self.detected_features.features:
            if feature.type not in self.logged_detection_types:
                logging.debug(
                    f"Feature | {feature.type} | {True}"
                )

        event.accept()


def main():
    from fibsem.detection.detection import Feature, move_based_on_detection
    import fibsem.ui.windows as fibsem_ui_windows
    from fibsem import movement, acquire
    from liftout import actions 
    from fibsem.detection.detection import NeedleTip, ImageCentre


    microscope, settings = utils.setup_session(protocol_path=r"C:\Users\Admin\Github\autoliftout\liftout\protocol\protocol.yaml")
    
    app = QtWidgets.QApplication([])

    import random

    # beam_type = BeamType.ELECTRON
    features = [NeedleTip(), 
                ImageCentre()]

        
    for beam_type in [BeamType.ELECTRON, BeamType.ION]:

        settings.image.beam_type = beam_type
        settings.image.hfw = 400e-6

        # actions.move_needle_to_liftout_position(microscope)

        det = fibsem_ui_windows.detect_features_v2(microscope, settings, features, validate=True)

        print("features: ", det.features)
        print("distance: ", det.distance)
        print("feature 1 position: ", det.features[0].feature_m)

        # from liftout.patterning import MillingPattern
        # milling_pattern = MillingPattern.Weld
        # point = det.features[0].feature_m
        # change_pattern = True
        # auto_continue = False

        # settings.image.hfw = 80e-6

        # from liftout.autoliftout import milling_ui

        # milling_ui(microscope, settings, milling_pattern, point = point, change_pattern = change_pattern, auto_continue=auto_continue)


        # move back 5um x
        # det.distance.x += -5e-6

        # for eb needle move: positive = up
        # for ib needle move: positive = down
        # if beam_type is BeamType.ION:
        #     move_x = False
        # else:
        #     move_x = True
        # move_based_on_detection(microscope, settings, det, beam_type=beam_type, move_x=True)

        # acquire.take_reference_images(microscope, settings.image)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()




# import logging
# from copy import deepcopy
# from pathlib import Path

# import napari
# import numpy as np
# from PyQt5.QtWidgets import QMessageBox

# from fibsem import acquire, conversions, validation
# from fibsem.acquire import take_reference_images
# from fibsem.detection import detection
# from fibsem.detection.detection import (
#     DetectedFeatures,
#     Feature,
#     ImageCentre,
#     LamellaCentre,
#     LamellaLeftEdge,
#     LamellaRightEdge,
#     LandingPost,
#     NeedleTip,
#     detect_features,
# )
# from fibsem.microscope import FibsemMicroscope
# from fibsem.segmentation.model import load_model
# from fibsem.structures import BeamType, FibsemStagePosition, MicroscopeSettings, Point
# from fibsem.ui import utils as fibsem_ui
# from fibsem.ui.FibsemDetectionUI import FibsemDetectionUI
# from fibsem.ui.FibsemDetectionWidget import detection_ui
# from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
# from fibsem.ui.FibsemManipulatorWidget import FibsemManipulatorWidget
# from fibsem.ui.FibsemMillingWidget import FibsemMillingWidget
# from fibsem.ui.utils import message_box_ui


# def detect_features_v2(
#     microscope: FibsemMicroscope,
#     settings: MicroscopeSettings,
#     features: tuple[Feature],
#     validate: bool = True,
# ) -> DetectedFeatures:
#     if settings.image.reduced_area is not None:
#         logging.info(
#             f"Reduced area is not compatible with model detection, disabling..."
#         )
#         settings.image.reduced_area = None

#     # take new image
#     image = acquire.new_image(microscope, settings.image)

#     # load model
#     ml_protocol = settings.protocol.get("ml", {})
#     checkpoint = ml_protocol.get("weights", None)
#     encoder = ml_protocol.get("encoder", "resnet18")
#     num_classes = int(ml_protocol.get("num_classes", 3))
#     cuda = ml_protocol.get("cuda", False)
#     model = load_model(checkpoint=checkpoint, encoder=encoder, nc=num_classes)

#     det = detection_ui(image=image, model=model, features=features, validate=validate)

#     # calculate features in microscope image coords
#     hfw = settings.image.hfw
#     pixelsize = hfw / settings.image.resolution[0]
#     # pixelsize = image.metadata.pixel_size.x
#     det.features[0].feature_m = conversions.image_to_microscope_image_coordinates(
#         det.features[0].px, image.data, pixelsize
#     )
#     det.features[1].feature_m = conversions.image_to_microscope_image_coordinates(
#         det.features[1].px, image.data, pixelsize
#     )

#     return det

# def run_validation_ui(
#     microscope: FibsemMicroscope, settings: MicroscopeSettings, log_path: Path
# ):
#     """Run validation checks to confirm microscope state before run."""

#     response = fibsem_ui.message_box_ui(
#         title="Microscope State Validation",
#         text="Do you want to validate the microscope state?",
#     )

#     if response is False:
#         logging.info(f"PRE_RUN_VALIDATION cancelled by user.")
#         return

#     logging.info(f"INIT | PRE_RUN_VALIDATION | STARTED")

#     # run validation
#     validation.validate_initial_microscope_state(microscope, settings)

#     # validate user configuration
#     validation._validate_configuration_values(microscope, settings.protocol)

#     # reminders
#     reminder_str = """Please check that the following steps have been completed:
#     \n - Experiment is inserted
#     \n - Confirm Operating Temperature
#     \n - Needle Calibration
#     \n - Ion Column Calibration
#     \n - Crossover Calibration
#     \n - Plasma Gas Valve Open
#     \n - Initial Grid and Landing Positions
#     """

#     response = fibsem_ui.message_box_ui(
#         title="Initialisation Checklist",
#         text=reminder_str,
#         buttons=QMessageBox.Ok,
#     )

#     # Loop backwards through the log, until we find the start of validation
#     with open(log_path) as f:
#         lines = f.read().splitlines()
#         validation_warnings = []
#         for line in lines[::-1]:
#             if "PRE_RUN_VALIDATION" in line:
#                 break
#             if "WARNING" in line:
#                 logging.info(line)
#                 validation_warnings.append(line)
#         logging.info(
#             f"{len(validation_warnings)} warnings were identified during intial setup."
#         )

#     if validation_warnings:
#         warning_str = f"The following {len(validation_warnings)} warnings were identified during initialisation."

#         for warning in validation_warnings[::-1]:
#             warning_str += f"\n{warning.split('—')[-1]}"

#         fibsem_ui.message_box_ui(
#             title="Initialisation Warning",
#             text=warning_str,
#             buttons=QMessageBox.Ok,
#         )

#     logging.info(f"INIT | PRE_RUN_VALIDATION | FINISHED")



import argparse
from fibsem.segmentation._nnunet import convert_to_nnunet_dataset

# Convert a fibsem dataset to nnunet format

def main():
    parser = argparse.ArgumentParser(description="Convert a dataset to nnunet format")
    parser.add_argument("--data_path", type=str, help="path to data")
    parser.add_argument("--label_path", type=str, help="path to labels")
    parser.add_argument("--nnunet_data_path", type=str, help="path to nnunet data")
    parser.add_argument(
        "--label_map", type=str, help="label map", required=False, default=None
    )
    parser.add_argument(
        "--filetype", type=str, help="filetype", required=False, default=".tif"
    )

    args = parser.parse_args()

    if args.label_map is not None:
        print(args.label_map)
        # open text file, read each line, and add to list of labels, remove newline character
        with open(args.label_map, "r") as f:
            labels = [line.rstrip("\n") for line in f]
        args.label_map = labels

    # print(args.data_path)
    # print(args.label_path)
    # print(args.nnunet_data_path)
    # print(args.label_map)
    # print(args.filetype)

    # return
    convert_to_nnunet_dataset(
        data_path=args.data_path,
        label_path=args.label_path,
        nnunet_data_path=args.nnunet_data_path,
        label_map=args.label_map,
        filetype=args.filetype,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(378, 531)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setDocumentMode(False)
        self.tabWidget.setTabsClosable(False)
        self.tabWidget.setTabBarAutoHide(False)
        self.tabWidget.setObjectName("tabWidget")
        self.gridLayout.addWidget(self.tabWidget, 1, 0, 1, 1)
        self.label_title = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setPointSize(16)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 2, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 378, 22))
        self.menubar.setObjectName("menubar")
        self.menuTools = QtWidgets.QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionCurrent_alignment = QtWidgets.QAction(MainWindow)
        self.actionCurrent_alignment.setObjectName("actionCurrent_alignment")
        self.actionManipulator_Positions_Calibration = QtWidgets.QAction(MainWindow)
        self.actionManipulator_Positions_Calibration.setObjectName("actionManipulator_Positions_Calibration")
        self.actionMinimap = QtWidgets.QAction(MainWindow)
        self.actionMinimap.setObjectName("actionMinimap")
        self.actionOpen_Minimap = QtWidgets.QAction(MainWindow)
        self.actionOpen_Minimap.setObjectName("actionOpen_Minimap")
        self.actionImage_Viewer = QtWidgets.QAction(MainWindow)
        self.actionImage_Viewer.setObjectName("actionImage_Viewer")
        self.menuTools.addAction(self.actionManipulator_Positions_Calibration)
        self.menuTools.addAction(self.actionOpen_Minimap)
        self.menubar.addAction(self.menuTools.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_title.setText(_translate("MainWindow", "OpenFIBSEM"))
        self.menuTools.setTitle(_translate("MainWindow", "Tools"))
        self.actionCurrent_alignment.setText(_translate("MainWindow", "Beam Current Alignment"))
        self.actionManipulator_Positions_Calibration.setText(_translate("MainWindow", "Manipulator Callibration"))
        self.actionMinimap.setText(_translate("MainWindow", "Minimap"))
        self.actionOpen_Minimap.setText(_translate("MainWindow", "Open Minimap"))
        self.actionImage_Viewer.setText(_translate("MainWindow", "Image Viewer"))





import os
import numpy as np
import onnx
import onnxruntime

from fibsem.segmentation.utils import decode_segmap_v2, download_checkpoint


### ONNX


class SegmentationModelONNX:

    def __init__(self, checkpoint: str = None):
        if checkpoint is not None:
            self.load_model(checkpoint)

    def load_model(self, checkpoint="autolamella-mega.onnx"):

        # download checkpoint if needed
        checkpoint = download_checkpoint(checkpoint)
        self.checkpoint = os.path.basename(checkpoint)

        # load inference session
        self.session = onnxruntime.InferenceSession(checkpoint, providers=["CPUExecutionProvider"])


    def inference(self, img: np.ndarray, rgb: bool = True):
        # preprocess
        imgt = img.astype(np.float32)
        imgt = np.expand_dims(imgt, axis=0)
        imgt = np.expand_dims(imgt, axis=0)
        imgt /= 255.0

        # inference
        ort_inputs = {self.session.get_inputs()[0].name: imgt}
        ort_outs = self.session.run(None, ort_inputs)


        # softmax
        outputs = ort_outs[0] # TODO: support batch size > 1
        outputs = np.exp(outputs) / np.sum(np.exp(outputs), axis=1, keepdims=True)
        masks = np.argmax(outputs, axis=1)
        mask = masks[0, :, :]

        # convert to rgb
        if rgb: 
            mask = decode_segmap_v2(mask)
        return mask

def export_model_to_onnx(checkpoint: str, onnx_path: str):
    
    from fibsem.segmentation.model import load_model
    import torch

    # get fibsem model
    model = load_model(checkpoint)
    model.model.to("cpu")
    model.model.eval()

    # From https://pytorch.org/tutorials/advanced/super_resolution_with_onnxruntime.html
    batch_size = 1
    x = torch.randn(batch_size, 1, 1024, 1536, requires_grad=True, dtype=torch.float)
    torch_out = model.model(x)

    # Export the model
    torch.onnx.export(model.model,               # model being run
                    x,                         # model input (or a tuple for multiple inputs)
                    onnx_path,   # where to save the model (can be a file or file-like object)
                    export_params=True,        # store the trained parameter weights inside the model file
                    do_constant_folding=True,  # whether to execute constant folding for optimization
                    input_names = ['input'],   # the model's input names
                    output_names = ['output'], # the model's output names
                    dynamic_axes={'input' : {0 : 'batch_size'},    # variable length axes
                                    'output' : {0 : 'batch_size'}})


    # load onnx model
    onnx_model = onnx.load(onnx_path)
    onnx.checker.check_model(onnx_model)

    # load inference session
    ort_session = onnxruntime.InferenceSession(onnx_path, providers=["CPUExecutionProvider"])

    def to_numpy(tensor):
        return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()

    # compute ONNX Runtime output prediction
    ort_inputs = {ort_session.get_inputs()[0].name: to_numpy(x)}
    ort_outs = ort_session.run(None, ort_inputs)

    # compare ONNX Runtime and PyTorch results
    np.testing.assert_allclose(to_numpy(torch_out), ort_outs[0], rtol=1e-03, atol=1e-05)

    print("Exported model has been tested with ONNXRuntime, and the result looks good!")

# export_model_to_onnx("autolamella-mega-latest.pt", "autolamella-mega-20231230.onnx")



import yaml
import os
from fibsem.config import CONFIG_PATH, DEFAULT_CONFIGURATION_VALUES



def get_column_tilt(manufacturer: str, beam: str) -> int:

    beam = f"{beam}-column-tilt"
    if manufacturer not in DEFAULT_CONFIGURATION_VALUES:
        raise ValueError(f"Unknown manufacturer: {manufacturer}")

    if beam not in DEFAULT_CONFIGURATION_VALUES[manufacturer]:
        raise ValueError(f"Unknown beam: {beam}")

    return DEFAULT_CONFIGURATION_VALUES[manufacturer][beam]
    

def generate_configuration(user_config: dict) -> dict:
    # load yaml
    with open(os.path.join(CONFIG_PATH, "microscope-configuration.yaml"), "r") as f:
        config = yaml.safe_load(f)

    # microscope
    config["info"]["name"] = user_config["name"]
    config["info"]["ip_address"] = user_config["ip_address"]
    config["info"]["manufacturer"] = user_config["manufacturer"]

    # stage
    config["stage"]["rotation_reference"] = user_config["rotation-reference"]
    config["stage"]["rotation_180"] = user_config["rotation-reference"] + 180
    config["stage"]["shuttle_pre_tilt"] = user_config["shuttle-pre-tilt"]

    # electron
    config["electron"]["eucentric_height"] = user_config["electron-beam-eucentric-height"]
    config["electron"]["column_tilt"] = get_column_tilt(config["info"]["manufacturer"], "electron")

    # ion
    config["ion"]["eucentric_height"] = user_config["ion-beam-eucentric-height"]
    config["ion"]["column_tilt"] = get_column_tilt(config["info"]["manufacturer"], "ion")

    return config


def get_user_config() -> dict:

    # enter name
    name = input("Enter configuration name: ")
    

    # enter ip address
    ip_address = input("Enter microscope PC IP address: ")
    assert ip_address.count(".") == 3, "IP address must be in the form: XXX.XXX.XXX.XXX"

    # enter manufacturer
    manufacturer = input("Enter microscope manufacturer (Thermo, Tescan, Demo): ")
    assert manufacturer in ["Thermo", "Tescan", "Demo"], "Unknown manufacturer. Must be Thermo, Tescan or Demo"

    # enter rotation reference
    rotation_reference = input("Enter rotation reference (degrees): ")
    rotation_reference = float(rotation_reference)
    # enter shuttle pre-tilt
    shuttle_pre_tilt = input("Enter shuttle pre-tilt (degrees): ")
    shuttle_pre_tilt = float(shuttle_pre_tilt)

    # enter electron beam eucentric height
    electron_beam_eucentric_height = input("Enter electron beam eucentric height (metres): ")
    electron_beam_eucentric_height = float(electron_beam_eucentric_height)

    # enter ion beam eucentric height
    ion_beam_eucentric_height = input("Enter ion beam eucentric height (metres): ")
    ion_beam_eucentric_height = float(ion_beam_eucentric_height)
    
    # generate configuration
    user_config = {
        "name":                           name,                          # a descriptive name for your configuration 
        "ip_address":                     ip_address,                    # the ip address of the microscope PC
        "manufacturer":                   manufacturer,                  # the microscope manufactuer, Thermo, Tescan or Demo                       
        "rotation-reference":             rotation_reference,            # the reference rotation value (rotation when loading)  [degrees]
        "shuttle-pre-tilt":               shuttle_pre_tilt,              # the pre-tilt of the shuttle                           [degrees]
        "electron-beam-eucentric-height": electron_beam_eucentric_height,# the eucentric height of the electron beam             [metres]
        "ion-beam-eucentric-height":      ion_beam_eucentric_height,     # the eucentric height of the ion beam                  [metres]
    }
    return user_config

def save_configuration(config: dict, path: str):
    """Save a configuration to a yaml file."""
    
    # save yaml
    filename = os.path.join(path, f"{config['info']['name']}.yaml")
    with open(filename, "w") as f:
        yaml.dump(config, f)

    return filename

import argparse
def gen_config_cli(path: str = None):
    """Generate a configuration from the command line."""
    
    parser = argparse.ArgumentParser(description="Generate a configuration from the command line.")
    parser.add_argument("--path", type=str, help="The path to save the configuration to.")
    args = parser.parse_args()
    
    if path is None:
        path = args.path

    # check path exists
    if not os.path.exists(path):
        raise ValueError(f"Path does not exist: {path}")

    print(f"OpenFIBSEM Configuration Generator\n")

    user_config = get_user_config()
    
    print(f"\nGenerating configuration: {user_config['name']}")
    config = generate_configuration(user_config)
    print("Configuration generated successfully.")

    filename = save_configuration(config, path)
    print(f"Configuration saved to: {filename}")


if __name__ == "__main__":
    gen_config_cli()



from typing import Optional, List

import numpy as np
import torch

from fibsem.segmentation.utils import decode_segmap_v2

from pathlib import Path
from fibsem.segmentation import config as scfg
import os
from huggingface_hub import hf_hub_download


from fibsem.segmentation import _nnunet as nnunet

# TODO: actually implement this
from abc import ABC
def SegmentationModelBase(ABC):

    def __init__(self, checkpoint: str):
        pass 


    def load_model(self, checkpoint: str) -> SegmentationModelBase:
        """Load the model, and optionally load a checkpoint"""
        raise NotImplementedError
    
    def inference(self, img: np.ndarray, rgb: bool = True) -> np.ndarray:
        """Run model inference on the input image"""
        raise NotImplementedError
    
class SegmentationModelNNUnet:
    def __init__(
        self,
        checkpoint: str = None,
    ) -> None:
        super().__init__()

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.colormap = scfg.get_colormap()
       
        self.checkpoint = checkpoint
        if checkpoint is not None:
            self.load_model(checkpoint=checkpoint)

    def load_model(self, checkpoint: str) -> None:
        """Load the model, and optionally load a checkpoint"""
        
        self.model = nnunet.load_model(path=checkpoint, device=self.device)

        return self.model

    def pre_process(self, img: np.ndarray) -> np.ndarray:
        """Pre-process the image for model inference"""

        # convert to 4D
        if img.ndim == 2:
            img = img[np.newaxis, np.newaxis, :, :]
        elif img.ndim == 3:
            img = img[np.newaxis, :, :, :]
        elif img.ndim == 4:
            img = img[:, :, :, :]
        else:
            raise ValueError(f"Invalid image shape: {img.shape}")
        
        # TODO: also do dtype conversions
        if not isinstance(img.dtype, np.float32):
            img = img.astype(np.float32)

        return img

    def inference(self, img: np.ndarray, rgb: bool = True) -> np.ndarray:
        """Run model inference on the input image"""
        
        img = self.pre_process(img)
        
        masks, scores = nnunet.inference(self.model, img)
        
        # convert to rgb image
        if rgb:
            masks = decode_segmap_v2(masks[0], self.colormap) # 2d only
        else:
            if masks.ndim>=3:
                masks = masks[0] # return 2d
        return masks

    def postprocess(self, masks):
        """Convert the model output to a rgb class map"""
        if masks.ndim == 3:
            masks = masks[0]
        return decode_segmap_v2(masks, self.colormap)

     


#!/usr/bin/env python3

import glob

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset, SubsetRandomSampler
from torchvision import transforms
import dask.array as da
import os
import tifffile as tff
import random
from PIL import Image

# transformations
ROT_ANGLE = 15
PROB = 0.1

transformations_input = transforms.Compose(
    [

        transforms.RandomRotation(ROT_ANGLE),
        transforms.RandomHorizontalFlip(p=PROB),
        transforms.RandomVerticalFlip(p=PROB),
        transforms.RandomAutocontrast(p=PROB),
        transforms.RandomEqualize(p=PROB),
        transforms.GaussianBlur(kernel_size=3, sigma=(0.1, 2.0)),
        transforms.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
    ]
)

transformations_target = transforms.Compose(
    [
        transforms.RandomRotation(ROT_ANGLE),
        transforms.RandomHorizontalFlip(p=PROB),
        transforms.RandomVerticalFlip(p=PROB),

    ]
)


class SegmentationDataset(Dataset):
    def __init__(self, images, masks, num_classes: int, transforms_input=None, transforms_target=None):
        self.images = images
        self.masks = masks
        self.num_classes = num_classes
        self.transforms_input = transforms_input
        self.transforms_target = transforms_target

        assert len(self.images) == len(self.masks), "Images and masks are not the same length"
        assert self.transforms_target is not None if self.transforms_input is not None else True, "transforms_target must be provided if transforms_input is provided"

    def _set_seed(self, seed):
        random.seed(seed)
        torch.manual_seed(seed)

    def __getitem__(self, idx):


        image = np.asarray(self.images[idx])
        mask = np.asarray(self.masks[idx])

        # if mask > num_class, set to zero
        mask[mask >= self.num_classes] = 0
        # - the problem was ToTensor was destroying the class index for the labels (rounding them to 0-1)
        # need to to transformation manually
        # mask = torch.tensor(mask).unsqueeze(0)

        image = torch.tensor(image).unsqueeze(0) # TODO: validate this behaviour
        mask = torch.tensor(mask).unsqueeze(0)
        
    
        image_max_val = torch.iinfo(image.dtype).max
        if image.dtype not in [torch.uint8, torch.int16]:
            raise ValueError(f"Image dtype {image.dtype} not supported")

        if self.transforms_input:
            seed = random.randint(0, 1000)
            self._set_seed(seed)
            image = self.transforms_input(image)
            self._set_seed(seed)
            mask = self.transforms_target(mask)

        # convert image to float32 scaled between 0-1
        image = image.float() / image_max_val
                        
        return image, mask

    def __len__(self):
        return len(self.images)

from pathlib import Path

def load_dask_dataset_v2(data_paths: list[Path], label_paths: list[Path]):

    sorted_img_filenames, sorted_mask_filenames = [], []
    for data_path, label_path in zip(data_paths, label_paths):
        sorted_img_filenames += sorted(glob.glob(os.path.join(data_path, "*.tif*")))
        sorted_mask_filenames += sorted(glob.glob(os.path.join(label_path, "*.tif*")))

    # TODO: change to dask-image
    img_arr = tff.imread(sorted_img_filenames, aszarr=True)
    mask_arr = tff.imread(sorted_mask_filenames, aszarr=True)

    images = da.from_zarr(img_arr)
    masks = da.from_zarr(mask_arr)

    images = images.rechunk(chunks=(1, images.shape[1], images.shape[2]))
    masks = masks.rechunk(chunks=(1, images.shape[1], images.shape[2]))

    return images, masks


def preprocess_data(data_paths: list[Path], label_paths: list[Path], num_classes: int = 3, 
                    batch_size: int = 1, val_split: float = 0.15, 
                    _validate_dataset:bool = True, apply_transforms: bool = False):
    
    # if _validate_dataset:
        # validate_dataset(data_path, label_path)

    if not isinstance(data_paths, list):
        data_paths = [data_paths]
    if not isinstance(label_paths, list):
        label_paths = [label_paths]

    images, masks = load_dask_dataset_v2(data_paths, label_paths)


    print(f"Loading dataset from {len(data_paths)} paths: {images.shape[0]}")
    for path in data_paths:
        print(f"Loaded from {path}")

    if apply_transforms is False:
        transformations_input, transformations_target = None, None

    # load dataset
    seg_dataset = SegmentationDataset(
        images, masks, num_classes, 
        transforms_input=transformations_input, 
        transforms_target=transformations_target
    )

    # train/validation splits
    dataset_size = len(seg_dataset)
    dataset_idx = list(range(dataset_size))
    split_idx = int(np.floor(val_split * dataset_size))
    train_idx = dataset_idx[split_idx:]
    val_idx = dataset_idx[:split_idx]

    train_sampler = SubsetRandomSampler(train_idx)
    val_sampler = SubsetRandomSampler(val_idx)

    train_data_loader = DataLoader(
        seg_dataset, batch_size=batch_size, sampler=train_sampler, drop_last=True
    )  # shuffle=True,
    print(f"Train dataset has {len(train_data_loader)} batches of size {batch_size}")

    # TODO: try larger val batch size? TODO: turn off transformations for validation set
    val_data_loader = DataLoader(
        seg_dataset, batch_size=1, sampler=val_sampler
    )  # shuffle=True,
    print(f"Validation dataset has {len(val_data_loader)} batches of size 1")

    return train_data_loader, val_data_loader


# ref: https://towardsdatascience.com/pytorch-basics-sampling-samplers-2a0f29f0bf2a

# Helper functions
def validate_dataset(data_path: str, label_path: str):
    print("validating dataset...")
    # get data
    filenames = sorted(glob.glob(os.path.join(data_path, "*.tif*")))
    labels = sorted(glob.glob(os.path.join(label_path,  "*.tif*")))

    # check length
    assert len(filenames) == len(labels), "Images and labels are not the same length"
    print(f"{len(filenames)} images, {len(labels)} labels")

    base_shape = tff.imread(filenames[0]).shape
    for i, (fname, lfname) in enumerate(list(zip(filenames, labels))):
        img, label = tff.imread(fname), tff.imread(lfname)

        if (img.shape[0:2] != label.shape[0:2]) or (img.shape[0:2] != base_shape[0:2]):

            raise ValueError(
                "invalid data, image shape is different to label shape",
                i,
                os.path.basename(fname),
                os.path.basename(lfname),
                img.shape,
                label.shape,
                "\n",
                "You can run convert_img_size() in utils.py to convert all images and labels in the dataset to the desired size.",
            )

        if (img.ndim > 2) or (label.ndim > 2):

            raise ValueError(
                "Image has too many dimensions, must be in 2D grayscale format.",
                i,
                os.path.basename(fname),
                os.path.basename(lfname),
                img.shape,
                label.shape,
                "\n",
                "You can run convert_to_grayscale() in utils.py to convert all images and labels in the dataset to the desired format.",
            )

        if (img.shape[0] % 32 != 0) or (img.shape[1] % 32 != 0):
            raise ValueError(
                "Wrong padding, dimensions must be divisible by 32.",
                i,
                os.path.basename(fname),
                os.path.basename(lfname),
                img.shape,
                label.shape,
                "\n",
                "You can run pad_data in utils.py to convert the dataset to correct format.",
            )

    print("finished validating dataset.")



import logging
import traceback
import yaml
import os 
import napari
import napari.utils.notifications
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal

import fibsem
from fibsem import config as cfg
from fibsem import utils
from fibsem.microscope import FibsemMicroscope
from fibsem.structures import MicroscopeSettings, SystemSettings
from fibsem.ui.qtdesigner_files import FibsemSystemSetupWidget
from fibsem.ui.utils import _get_file_ui, _get_save_file_ui
from fibsem.ui import _stylesheets

def log_status_message(step: str):
    logging.debug(
        f"STATUS | System Widget | {step}"
    )


class FibsemSystemSetupWidget(FibsemSystemSetupWidget.Ui_Form, QtWidgets.QWidget):
    connected_signal = pyqtSignal()
    disconnected_signal = pyqtSignal()

    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        viewer: napari.Viewer = None,
        parent=None,
    ):
        super(FibsemSystemSetupWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.parent = parent
        self.microscope = microscope
        self.settings = settings
        self.viewer = viewer

        self.setup_connections()
        
        self.update_ui()

    def setup_connections(self):

        # connection
        self.pushButton_connect_to_microscope.clicked.connect(self.connect_to_microscope)
        
        # configuration   
        self.comboBox_configuration.addItems(cfg.USER_CONFIGURATIONS.keys())
        self.comboBox_configuration.setCurrentText(cfg.DEFAULT_CONFIGURATION_NAME) 
        self.comboBox_configuration.currentTextChanged.connect(lambda: self.load_configuration(None))
        self.toolButton_import_configuration.clicked.connect(self.import_configuration_from_file)
    
        self.pushButton_apply_configuration.clicked.connect(lambda: self.apply_microscope_configuration(None))
        self.pushButton_apply_configuration.setToolTip(f"Apply configuration can take some time. Please make sure the microscope beams are both on.")

    def load_configuration(self, configuration_name: str):
        if configuration_name is None:
            configuration_name = self.comboBox_configuration.currentText()
        
        configuration_path = cfg.USER_CONFIGURATIONS[configuration_name]["path"]

        if configuration_path is None:  
            napari.utils.notifications.show_error(f"Configuration {configuration_name} not found.")
            return
    
        # load the configuration
        self.settings = utils.load_microscope_configuration(configuration_path)

        from pprint import pprint 

        pprint(self.settings.to_dict()["info"])

        return configuration_path
    
    def import_configuration_from_file(self):
    
        path = _get_file_ui(msg="Select microscope configuration file", 
            path=cfg.CONFIG_PATH, _filter="YAML (*.yaml *.yml)")

        if path == "":
            napari.utils.notifications.show_error(f"No file selected. Configuration not loaded.")
            return
        
        # TODO: validate configuration  

        # ask user to add to user configurations
        configuration_name = os.path.basename(path).removesuffix(".yaml")

        if configuration_name not in cfg.USER_CONFIGURATIONS: 
            from fibsem.ui.utils import message_box_ui
            msg = f"Would you like to add this configuration to the user configurations?"
            ret = message_box_ui(text=msg, title="Add to user configurations?")

            # add to user configurations
            if ret:
                cfg.add_configuration(configuration_name=configuration_name, path=path)
                
                # set default configuration
                msg = f"Would you like to make this the default configuration?"
                ret = message_box_ui(text=msg, title="Set default configuration?")
                
                if ret:
                    cfg.set_default_configuration(configuration_name=configuration_name)

        # add configuration to combobox
        self.comboBox_configuration.addItem(configuration_name)
        self.comboBox_configuration.setCurrentText(configuration_name)

    def connect_to_microscope(self):
                
        _microscope_connected = bool(self.microscope)

        if _microscope_connected:
            self.microscope.disconnect()
            self.microscope, self.settings = None, None
        else:

            napari.utils.notifications.show_info(f"Connecting to microscope...")

            configuration_path = self.load_configuration(None)

            if configuration_path is None:
                napari.utils.notifications.show_error(f"Configuration not selected.")
                return

            # connect
            self.microscope, self.settings = utils.setup_session(
                config_path=configuration_path,
            )

            # user notification
            msg = f"Connected to microscope at {self.microscope.system.info.ip_address}"
            logging.info(msg)
            napari.utils.notifications.show_info(msg)

        self.update_ui()
            

    def apply_microscope_configuration(self, system_settings: SystemSettings = None):
        """Apply the microscope configuration to the microscope."""

        if self.microscope is None:
            napari.utils.notifications.show_error(f"Microscope not connected.")
            return
        
        # apply the configuration
        self.microscope.apply_configuration(system_settings=system_settings)
    



    # def apply_defaults_settings(self):
    #     microscope_settings = self.get_default_settings_from_ui()

    #     self.microscope.set_beam_system_settings(microscope_settings.system.ion)
    #     self.microscope.set_beam_system_settings(microscope_settings.system.electron)
        
    #     # TODO: complete this system setting
    #     if self.parent:
    #         if self.parent.image_widget:
    #             self.parent.image_widget.set_ui_from_settings(microscope_settings.image, beam_type=microscope_settings.image.beam_type)
    #         if self.parent.milling_widget:
    #             self.parent.milling_widget.set_milling_settings_ui(microscope_settings.milling)

    #     self.get_stage_settings_from_ui()
    #     self.get_model_from_ui()

    def update_ui(self):

        _microscope_connected = bool(self.microscope)
        self.pushButton_apply_configuration.setVisible(_microscope_connected)
        self.pushButton_apply_configuration.setEnabled(_microscope_connected and cfg._APPLY_CONFIGURATION_ENABLED)

        if _microscope_connected:
            self.pushButton_connect_to_microscope.setText("Microscope Connected")
            self.pushButton_connect_to_microscope.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_apply_configuration.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.connected_signal.emit()
            
            info = self.microscope.system.info
            self.label_connection_information.setText(f"Connected to {info.manufacturer}-{info.model} at {info.ip_address}")

        else:
            self.pushButton_connect_to_microscope.setText("Connect To Microscope")
            self.pushButton_connect_to_microscope.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)
            self.pushButton_apply_configuration.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)
            self.disconnected_signal.emit()
            self.label_connection_information.setText("Not connected to microscope")


def main():

    viewer = napari.Viewer(ndisplay=2)
    movement_widget = FibsemSystemSetupWidget()
    viewer.window.add_dock_widget(
        movement_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()



import numpy as np
from fibsem.structures import Point, FibsemImage


def image_to_microscope_image_coordinates(
    coord: Point, image: np.ndarray, pixelsize: float
) -> Point:
    """
    Convert an image pixel coordinate to a microscope image coordinate.

    The microscope image coordinate system is centered on the image with positive Y-axis pointing upwards.

    Args:
        coord (Point): A Point object representing the pixel coordinates in the original image.
        image (np.ndarray): A numpy array representing the image.
        pixelsize (float): The pixel size in meters.

    Returns:
        Point: A Point object representing the corresponding microscope image coordinates in meters.
    """
    # convert from image pixel coord (0, 0) top left to microscope image (0, 0) mid

    # shape
    cy, cx = np.asarray(image.shape) // 2

    # distance from centre?
    dy = float(-(coord.y - cy))  # neg = down
    dx = float(coord.x - cx)  # neg = left

    point_m = convert_point_from_pixel_to_metres(Point(dx, dy), pixelsize)

    return point_m


def get_lamella_size_in_pixels(
    img: FibsemImage, protocol: dict, use_trench_height: bool = False
) -> tuple[int]:
    """Get the relative size of the lamella in pixels based on the hfw of the image.

    Args:
        img (FibsemImage): A reference image.
        protocol (dict): A dictionary containing the protocol information.
        use_trench_height (bool, optional): If True, returns the height of the trench instead of the lamella. Default is False.

    Returns:
        tuple[int]: A tuple containing the height and width of the lamella in pixels.
    """
    # get real size from protocol
    lamella_width = protocol["lamella_width"]
    lamella_height = protocol["lamella_height"]

    total_height = lamella_height
    if use_trench_height:
        trench_height = protocol["stages"][0]["trench_height"]
        total_height += 2 * trench_height

    # convert to m
    pixelsize = img.metadata.pixel_size.x
    width, height = img.metadata.image_settings.resolution
    vfw = convert_pixels_to_metres(height, pixelsize)
    hfw = convert_pixels_to_metres(width, pixelsize)

    # lamella size in px (% of image)
    lamella_height_px = int((total_height / vfw) * height)
    lamella_width_px = int((lamella_width / hfw) * width)

    return (lamella_height_px, lamella_width_px)


def convert_metres_to_pixels(distance: float, pixelsize: float) -> int:
    """
    Convert a distance in metres to pixels based on a given pixel size.

    Args:
        distance (float): The distance to convert, in metres.
        pixelsize (float): The size of a pixel in metres.

    Returns:
        int: The distance converted to pixels, as an integer.
    """
    return int(distance / pixelsize)


def convert_pixels_to_metres(pixels: int, pixelsize: float) -> float:
    """
    Convert a distance in pixels to metres based on a given pixel size.

    Args:
        pixels (int): The number of pixels to convert.
        pixelsize (float): The size of a pixel in metres.

    Returns:
        float: The distance converted to metres.
    """
    return float(pixels * pixelsize)


def distance_between_points(p1: Point, p2: Point) -> Point:
    """
    Calculate the Euclidean distance between two points, returning a Point object representing the result.

    Args:
        p1 (Point): The first point.
        p2 (Point): The second point.

    Returns:
        Point: A Point object representing the distance between the two points.
    """

    return Point(x=(p2.x - p1.x), y=(p2.y - p1.y))


def convert_point_from_pixel_to_metres(point: Point, pixelsize: float) -> Point:
    """
    Convert a Point object from pixel coordinates to metre coordinates, based on a given pixel size.

    Args:
        point (Point): The Point object to convert.
        pixelsize (float): The size of a pixel in metres.

    Returns:
        Point: The converted Point object, with its x and y values in metre coordinates.
    """
    point_m = Point(
        x=convert_pixels_to_metres(point.x, pixelsize),
        y=convert_pixels_to_metres(point.y, pixelsize),
    )

    return point_m


def convert_point_from_metres_to_pixel(point: Point, pixelsize: float) -> Point:
    """
    Convert a Point object from metre coordinates to pixel coordinates, based on a given pixel size.

    Args:
        point (Point): The Point object to convert.
        pixelsize (float): The size of a pixel in metres.

    Returns:
        Point: The converted Point object, with its x and y values in pixel coordinates.
    """
    point_px = Point(
        x=convert_metres_to_pixels(point.x, pixelsize),
        y=convert_metres_to_pixels(point.y, pixelsize),
    )
    return point_px



import glob
import logging
import os
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from skimage import transform

from fibsem import acquire, conversions, utils
from fibsem.structures import (BeamType, FibsemImage, FibsemImageMetadata, MicroscopeSettings,
                               FibsemStagePosition, Point)
from fibsem.microscope import FibsemMicroscope
import time

def _tile_image_collection(microscope: FibsemMicroscope, settings: MicroscopeSettings, grid_size:float, 
    tile_size:float, overlap: float = 0.0, cryo: bool = True, parent_ui = None) -> dict: 

    # TODO: OVERLAP + STITCH
    n_rows, n_cols = int(grid_size // tile_size), int(grid_size // tile_size)
    dx, dy = settings.image.hfw, settings.image.hfw 

    dy *= -1 # need to invert y-axis

    # fixed image settings
    settings.image.autogamma = False

    logging.info(f"TILE COLLECTION: {settings.image.filename}")
    logging.info(f"Taking n_rows={n_rows}, n_cols={n_cols} ({n_rows*n_cols}) images. Grid Size = {grid_size*1e6} um, Tile Size = {tile_size*1e6} um")
    logging.info(f"dx: {dx*1e6} um, dy: {dy*1e6} um")

    # start in the middle of the grid

    start_state = microscope.get_microscope_state()
    
    # we save all intermediates into a folder with the same name as the filename, then save the stitched image into the parent folder
    prev_path = settings.image.path
    settings.image.path = os.path.join(settings.image.path, settings.image.filename)
    os.makedirs(settings.image.path, exist_ok=True)
    prev_label = settings.image.filename
    
    # BIG_IMAGE FOR DEBUGGING ONLY
    settings.image.hfw = grid_size
    settings.image.filename = "big_image"
    big_image = acquire.new_image(microscope, settings.image)
    
    # TOP LEFT CORNER START
    settings.image.hfw = tile_size
    settings.image.filename = prev_label
    settings.image.autocontrast = False # required for cryo
    start_move = grid_size / 2 - tile_size / 2
    dxg, dyg = start_move, start_move
    dyg *= -1

    microscope.stable_move(dx=-dxg, dy=-dyg, beam_type=settings.image.beam_type, static_wd=True)
    state = microscope.get_microscope_state()
    images = []

    # stitched image
    shape = settings.image.resolution
    arr = np.zeros(shape=(n_rows*shape[0], n_cols*shape[1]), dtype=np.uint8)
    _counter = 0
    _total = n_rows*n_cols
    for i in range(n_rows):

        microscope.safe_absolute_stage_movement(state.stage_position)
        
        img_row = []
        microscope.stable_move(
            dx=0,
            dy=i*dy, 
            beam_type=settings.image.beam_type, 
            static_wd=True)


        for j in range(n_cols):
            settings.image.filename = f"tile_{i}_{j}"
            microscope.stable_move(dx=dx*(j!=0),  dy=0, beam_type=settings.image.beam_type) # dont move on the first tile?

            logging.info(f"ACQUIRING IMAGE {i}, {j}")
            image = acquire.new_image(microscope, settings.image)

            # stitch image
            arr[i*shape[0]:(i+1)*shape[0], j*shape[1]:(j+1)*shape[1]] = image.data
            
            if parent_ui:
                _counter+=1 
                parent_ui._update_tile_collection.emit({"msg": "Tile Collected", "i": i, "j": j, 
                    "n_rows": n_rows, "n_cols": n_cols, "image": arr, "counter": _counter, "total": _total })
                time.sleep(1)

            img_row.append(image)
        images.append(img_row)

    # restore initial state
    microscope.set_microscope_state(start_state)
    settings.image.path = prev_path

    ddict = {"grid_size": grid_size, "tile_size": tile_size, "n_rows": n_rows, "n_cols": n_cols, 
            "image_settings": settings.image, 
            "dx": dx, "dy": dy, "cryo": cryo,
            "start_state": start_state, "prev-filename": prev_label, "start_move": start_move, "dxg": dxg, "dyg": dyg,
            "images": images, "big_image": big_image, "stitched_image": arr}

    return ddict


# TODO: stitch while collecting
def _stitch_images(images, ddict: dict, overlap=0, parent_ui = None) -> FibsemImage:

    # arr = np.array(images)
    # n_rows, n_cols = arr.shape[0], arr.shape[1]
    # shape = arr[0, 0].data.shape

    # arr = np.zeros(shape=(n_rows*shape[0], n_cols*shape[1]), dtype=np.uint8)

    # for i in range(n_rows):
    #     for j in range(n_cols):
    #         arr[i*shape[0]:(i+1)*shape[0], j*shape[1]:(j+1)*shape[1]] = images[i][j].data
            
    #         if parent_ui:
    #             parent_ui._update_tile_collection.emit({"msg": "Tile Stitched", "i": i, "j": j, 
    #                 "n_rows": n_rows, "n_cols": n_cols, "image": arr })
    
    arr = ddict["stitched_image"]

    # convert to fibsem image
    image = FibsemImage(data=arr, metadata=images[0][0].metadata)
    image.metadata.microscope_state = deepcopy(ddict["start_state"])
    image.metadata.image_settings = ddict["image_settings"]
    image.metadata.image_settings.hfw = deepcopy(float(ddict["grid_size"]))
    image.metadata.image_settings.resolution = deepcopy([arr.shape[0], arr.shape[1]])

    filename = os.path.join(image.metadata.image_settings.path, f'{ddict["prev-filename"]}')
    image.save(filename)

    # for cryo need to histogram equalise
    if ddict.get("cryo", False):
        from fibsem.imaging.autogamma import auto_gamma
        image = auto_gamma(image, method="autogamma")

    filename = os.path.join(image.metadata.image_settings.path, f'{ddict["prev-filename"]}-autogamma')
    image.save(filename)

    # save ddict as yaml
    del ddict["images"]
    del ddict["big_image"]

    ddict["image_settings"] = ddict["image_settings"].to_dict()
    ddict["start_state"] = ddict["start_state"].to_dict()
    filename = os.path.join(filename, f'{ddict["prev-filename"]}') # subdir
    utils.save_yaml(filename, ddict) 

    return image


def _tile_image_collection_stitch(microscope, settings, grid_size, tile_size, overlap=0, cryo:bool=True, parent_ui = None) -> FibsemImage:

    ddict = _tile_image_collection(microscope, settings, grid_size, tile_size, cryo=cryo, parent_ui=parent_ui)
    image = _stitch_images(ddict["images"], ddict, overlap=overlap, parent_ui=parent_ui)

    return image


def _stitch_arr(images:list[FibsemImage], dtype=np.uint8):

    arr = np.array(images)
    n_rows, n_cols = arr.shape[0], arr.shape[1]
    shape = arr[0, 0].data.shape

    arr = np.zeros(shape=(n_rows*shape[0], n_cols*shape[1]), dtype=dtype)

    for i in range(n_rows):
        for j in range(n_cols):
            arr[i*shape[0]:(i+1)*shape[0], j*shape[1]:(j+1)*shape[1]] = images[i][j].data  

    return arr



def _create_tiles(image: np.ndarray, n_rows, n_cols, tile_size, overlap=0):
    # create tiles
    tiles = []
    for i in range(n_rows):

        for j in range(n_cols):

            # get tile
            tile = image[i*tile_size:(i+1)*tile_size, j*tile_size:(j+1)*tile_size]

            # append to list
            tiles.append(tile)

    tiles = np.array(tiles)

    return tiles



def _calculate_repojection(image: FibsemImage, pos: FibsemStagePosition):

    # difference between current position and image position
    delta = pos - image.metadata.microscope_state.stage_position

    # projection of the positions onto the image
    dx = delta.x
    dy = np.sqrt(delta.y**2 + delta.z**2) # TODO: correct for perspective here
    dy = dy if (delta.y<0) else -dy

    pt_delta = Point(dx, dy)
    px_delta = pt_delta._to_pixels(image.metadata.pixel_size.x)

    image_centre = Point(x=image.data.shape[1]/2, y=image.data.shape[0]/2)
    point = image_centre + px_delta

    # NB: there is a small reprojection error that grows with distance from centre
    # print(f"ERROR: dy: {dy}, delta_y: {delta.y}, delta_z: {delta.z}")

    return point


def _reproject_positions(image:FibsemImage, positions: list[FibsemStagePosition], _bound: bool=False):
    # reprojection of positions onto image coordinates
    points = []
    for pos in positions:


        # hotfix (pat): demo returns None positions #240
        if image.metadata.microscope_state.stage_position.x is None:
            image.metadata.microscope_state.stage_position.x = 0
        if image.metadata.microscope_state.stage_position.y is None:
            image.metadata.microscope_state.stage_position.y = 0
        if image.metadata.microscope_state.stage_position.z is None:
            image.metadata.microscope_state.stage_position.z = 0
        if image.metadata.microscope_state.stage_position.r is None:
            image.metadata.microscope_state.stage_position.r = 0
        if image.metadata.microscope_state.stage_position.t is None:
            image.metadata.microscope_state.stage_position.t = 0      
                
        # automate logic for transforming positions
        # assume only two valid positions are when stage is flat to either beam...  
        # r needs to be 180 degrees different
        # currently only one way: Flat to Ion -> Flat to Electron
        dr = abs(np.rad2deg(image.metadata.microscope_state.stage_position.r - pos.r))
        # print("dr: ", dr)
        if np.isclose(dr, 180, atol=2):     
            # print("transforming position")
            pos = _transform_position(pos)

        pt = _calculate_repojection(image, pos)
        pt.name = pos.name
        
        if _bound:
            if pt.x<0 or pt.x>image.data.shape[1] or pt.y<0 or pt.y>image.data.shape[0]:
                continue 
        
        points.append(pt)
    
    return points




def _plot_positions(image: FibsemImage, positions: list[FibsemStagePosition], show:bool=False, minimap: bool=False, 
    _clip: bool=False, _bound: bool= True) -> plt.Figure:

    points = _reproject_positions(image, positions)

    # plot on matplotlib
    fig = plt.figure(figsize=(15, 15))
    plt.imshow(image.data, cmap="gray")

    COLOURS = ["lime", "blue", "cyan", "magenta", 
        "hotpink", "yellow", "orange", "red"]
    for i, (pos, pt) in enumerate(zip(positions, points)):

        # if points outside image, don't plot
        if _bound:
            if pt.x<0 or pt.x>image.data.shape[1] or pt.y<0 or pt.y>image.data.shape[0]:
                continue          
        
        # clip points to image 
        if _clip:
            pt.x = np.clip(pt.x, 0, image.data.shape[1])
            pt.y = np.clip(pt.y, 0, image.data.shape[0])


        c =COLOURS[i%len(COLOURS)]
        plt.plot(pt.x, pt.y, ms=20, c=c, marker="+", markeredgewidth=2, filename=f"{pos.name}")
        if minimap:
            fontsize = 30
        else:
            fontsize = 14
        # draw filename next to point
        plt.text(pt.x-225, pt.y-50, pos.name, fontsize=fontsize, color=c, alpha=0.75)

    plt.axis("off")
    if show:
        plt.show()

    return fig


# TODO: these probs should be somewhere else
def _convert_image_coord_to_position(microscope, settings, image, coords) -> FibsemStagePosition:

    # microscope coordinates
    point = conversions.image_to_microscope_image_coordinates(
        Point(x=coords[1], y=coords[0]), image.data, image.metadata.pixel_size.x,
    )

    _new_position = microscope.project_stable_move( 
            dx=point.x, dy=point.y, 
            beam_type=image.metadata.image_settings.beam_type, 
            base_position=image.metadata.microscope_state.stage_position)

    return _new_position


def _convert_image_coords_to_positions(microscope, settings, image, coords) -> list[FibsemStagePosition]:

    positions = []
    for i, coord in enumerate(coords):
        positions.append(_convert_image_coord_to_position(microscope, settings, image, coord))
        positions[i].name = f"Position {i:02d}"
    return positions


def _minimap(minimap_image: FibsemImage, positions: list[FibsemStagePosition]):
        import matplotlib.pyplot as plt
        from matplotlib.backends.backend_agg import FigureCanvasAgg
        from PIL import Image
        from PyQt5.QtGui import QImage, QPixmap
        pil_image = None

        import matplotlib.pyplot as plt
        plt.close("all")
        fig = _plot_positions(image=minimap_image, positions = positions, minimap=True)
        plt.tight_layout(pad=0)
        canvas = FigureCanvasAgg(fig)

        # Render the figure onto the canvas
        canvas.draw()

        # Get the RGBA buffer from the canvas
        buf = canvas.buffer_rgba()
        # Convert the buffer to a PIL Image
        pil_image = Image.frombuffer('RGBA', canvas.get_width_height(), buf, 'raw', 'RGBA', 0, 1)
        pil_image = pil_image.resize((300, 300), Image.ANTIALIAS)
        # Convert the PIL image to a QImage
        image_qt = QImage(pil_image.tobytes(), pil_image.width, pil_image.height, QImage.Format_RGBA8888)
        # Convert the QImage to a QPixmap 
        qpixmap = QPixmap.fromImage(image_qt)
        
        return qpixmap



##### THERMO ONLY

X_OFFSET = -0.0005127403888932854 
Y_OFFSET = 0.0007937916666666666

def _to_specimen_coordinate_system(pos: FibsemStagePosition):
    """Converts a position in the raw coordinate system to the specimen coordinate system"""

    specimen_offset = FibsemStagePosition(x=X_OFFSET, y=Y_OFFSET, z=0.0, r=0, t=0, coordinate_system="RAW")
    specimen_position = pos - specimen_offset

    return specimen_position

def _to_raw_coordinate_system(pos: FibsemStagePosition):
    """Converts a position in the raw coordinate system to the specimen coordinate system"""

    specimen_offset = FibsemStagePosition(x=X_OFFSET, y=Y_OFFSET, z=0.0, r=0, t=0, coordinate_system="RAW")
    raw_position = pos + specimen_offset

    return raw_position


def _transform_position(pos: FibsemStagePosition):
    """This function takes in a position flat to a beam, and outputs the position if stage was rotated / tilted flat to the other beam)"""

    specimen_position = _to_specimen_coordinate_system(pos)
    # print("raw      pos: ", pos)
    # print("specimen pos: ", specimen_position)

    # # inverse xy (rotate 180 degrees)
    specimen_position.x = -specimen_position.x
    specimen_position.y = -specimen_position.y

    # movement offset (calibration for compucentric rotation error)
    specimen_position.x += 50e-6
    specimen_position.y += 25e-6

    # print("rotated pos: ", specimen_position)

    # _to_raw_coordinates
    transformed_position = _to_raw_coordinate_system(specimen_position)
    transformed_position.name = pos.name

    # print("trans   pos: ", transformed_position)

    return transformed_position

from fibsem.structures import ImageSettings
def _update_image_region(microscope: FibsemMicroscope, image_settings: ImageSettings, image: FibsemImage, position: FibsemStagePosition) -> FibsemImage:
    
    region_image = acquire.new_image(microscope, image_settings)
    rows,cols = region_image.data.shape[0], region_image.data.shape[1]

    position_point = _reproject_positions(image, [position])[0]

    ymin = max(0, int(position_point.y)-rows//2)
    xmin = max(0, int(position_point.x)-cols//2)

    ymax = min(ymin+rows, image.data.shape[0])
    xmax = min(xmin+cols, image.data.shape[1])

    width = xmax - xmin
    height = ymax - ymin

    # overwrite the image with the region image, but only inside the image dimensions
    image.data[ymin:ymax, xmin:xmax] = region_image.data[:height, :width]

    return image


import argparse
import os
import logging
import napari
import napari.utils.notifications
import fibsem
from PyQt5 import QtWidgets
from fibsem import config as cfg
from fibsem import constants, conversions, utils
from fibsem.ui.qtdesigner_files import FibsemMicroscopeConfigurationWidgetBase
from fibsem.ui import _stylesheets
from fibsem.structures import BeamType
from fibsem.ui.utils import _get_file_ui, _get_save_file_ui, message_box_ui
from pprint import pprint

CONFIGURATION = {
    "tooltips": {"electron-column-tilt": "The tilt of the electron column. Typically 0 degrees.",
                 "ion-column-tilt": "The tilt of the ion column (Thermo = 52 deg, Tescan = 55 deg).", 
                 "rotation-reference": "The rotation of the stage when loading the sample (under the SEM)",
                 "rotation-180": "The rotation of the stage under the FIB (180 degrees from rotation-reference)",},
}


class FibsemMicroscopeConfigurationWidgetBase(FibsemMicroscopeConfigurationWidgetBase.Ui_Form, QtWidgets.QWidget):
    def __init__(
        self,
        path: str = None,
        viewer: napari.Viewer = None,
        parent=None,
    ):
        super(FibsemMicroscopeConfigurationWidgetBase, self).__init__(parent=parent)
        self.setupUi(self)
        self.setWindowTitle("Microscope Configuration")

        self.viewer = viewer

        self.setup_connections()

        if path is not None:
            self.path = path
            self.configuration = self.load_configuration(path)
            self.load_configuration_to_ui()

    def setup_connections(self):

        # combo boxes
        self.comboBox_configuration_manufacturer.addItems(cfg.__SUPPORTED_MANUFACTURERS__)
        self.comboBox_imaging_beam_type.addItems([b.name for b in BeamType])
        self.comboBox_configuration_manufacturer.currentTextChanged.connect(self.update_configuration_ui)

        # manufactuer specific widgets
        self.tescan_widgets = [
            self.label_milling_dwell_time,  self.doubleSpinBox_milling_dwell_time,
            self.label_milling_rate,        self.doubleSpinBox_milling_rate,
            self.label_milling_spot_size,   self.doubleSpinBox_milling_spot_size,
            self.label_milling_preset,      self.lineEdit_milling_preset,
        ]

        self.thermo_widgets = [
                self.label_milling_voltage, self.doubleSpinBox_milling_voltage,
                self.label_milling_current, self.doubleSpinBox_milling_current,
        ]

        # tooltips
        self.spinBox_electron_column_tilt.setToolTip(CONFIGURATION["tooltips"]["electron-column-tilt"])
        self.spinBox_ion_column_tilt.setToolTip(CONFIGURATION["tooltips"]["ion-column-tilt"])
        self.spinBox_stage_rotation_reference.setToolTip(CONFIGURATION["tooltips"]["rotation-reference"])
        self.spinBox_stage_rotation_180.setToolTip(CONFIGURATION["tooltips"]["rotation-180"])

        # dynamic changing configuration
        # change configuration values based on derived values, e.g. manufactuer -> column-tilt

        self.checkBox_show_advanced_settings.stateChanged.connect(self.toggle_advanced_settings)
        self.checkBox_show_advanced_settings.setVisible(False) # disable for now

    def toggle_advanced_settings(self):

        show_advanced = self.checkBox_show_advanced_settings.isChecked()

        for widget in self.advanced_widgets:
            widget.setVisible(show_advanced)


    def update_configuration_ui(self):

        manufactuer = self.comboBox_configuration_manufacturer.currentText()

        TESCAN_MANUFACTUER = manufactuer == "Tescan"
        THERMO_MANUFACTUER = manufactuer == "Thermo"
        DEMO_MANUFACTURER = manufactuer == "Demo"

        if DEMO_MANUFACTURER:
            TESCAN_MANUFACTUER, THERMO_MANUFACTUER = True, True
        
        # toggle widgets based on manufactuer
        for widget in self.tescan_widgets:
            widget.setVisible(TESCAN_MANUFACTUER)
        for widget in self.thermo_widgets:
            widget.setVisible(THERMO_MANUFACTUER)


        # set ion column tilts
        if TESCAN_MANUFACTUER:
            self.spinBox_ion_column_tilt.setValue(cfg.DEFAULT_CONFIGURATION_VALUES["Tescan"]["ion-column-tilt"])
        elif THERMO_MANUFACTUER:
            self.spinBox_ion_column_tilt.setValue(cfg.DEFAULT_CONFIGURATION_VALUES["Thermo"]["ion-column-tilt"])

    def save_configuration(self):

        configuration = self.read_configuration_from_ui()
        configuration_name = configuration["info"]["name"]
        path = os.path.join(os.path.dirname(self.path), f"{configuration_name}.yaml")

        # select save file 
        path = _get_save_file_ui(msg="Select microscope configuration file", 
                                path=path, 
                                _filter="YAML (*.yaml)",
                                parent=self)
        if path == "":
            napari.utils.notifications.show_error(f"Please select a file to save the configuration to. Configuration not saved.")
            return 
        
        # save configuration
        utils.save_yaml(path, configuration)
        napari.utils.notifications.show_info(f"Configuration {configuration_name} saved to {path}")

        # add to user configurations
        msg = f"Would you like to add this configuration ({configuration_name}) to the user configurations?"
        ret = message_box_ui(text=msg, title="Add to user configurations?")

        # add to user configurations
        if ret:
            cfg.add_configuration(configuration_name=configuration_name, path=path)
            
            # set default configuration
            msg = f"Would you like to make this the default configuration?"
            ret = message_box_ui(text=msg, title="Set default configuration?")
            
            if ret:
                cfg.set_default_configuration(configuration_name=configuration_name)

    def load_configuration_from_file(self):

        path = _get_file_ui(msg="Select microscope configuration file", 
                            path=cfg.DEFAULT_CONFIGURATION_PATH, 
                            _filter="YAML (*.yaml)", parent=self)

        if path == "":
            napari.utils.notifications.show_error(f"Please select a file to load the configuration from. Configuration not loaded.")
            return
        
        self.path = path
        self.configuration = self.load_configuration(path)
        self.load_configuration_to_ui()

    def load_configuration(self, path: str):
        
        self.configuration = utils.load_yaml(path)

        return self.configuration
    
    def load_configuration_to_ui(self):
        
        # core
        info_configuration = self.configuration["info"]
        self.lineEdit_configuration_name.setText(info_configuration["name"])
        self.lineEdit_configuration_path.setText(self.path)
        self.lineEdit_configuration_ip_address.setText(info_configuration["ip_address"])
        self.comboBox_configuration_manufacturer.setCurrentText(info_configuration["manufacturer"])

        # stage
        stage_configuration = self.configuration["stage"]
        self.spinBox_stage_rotation_reference.setValue(stage_configuration["rotation_reference"])
        self.spinBox_stage_rotation_180.setValue(stage_configuration["rotation_180"])
        self.spinBox_stage_shuttle_pre_tilt.setValue(int(stage_configuration["shuttle_pre_tilt"]))
        self.doubleSpinBox_stage_manipulator_height_limit.setValue(stage_configuration["manipulator_height_limit"] * constants.SI_TO_MILLI) # mm

        # electron
        electron_configuration = self.configuration["electron"]
        self.spinBox_electron_column_tilt.setValue(electron_configuration["column_tilt"])                               # deg
        self.doubleSpinBox_electron_eucentric_height.setValue(electron_configuration["eucentric_height"] * constants.SI_TO_MILLI) # mm
        self.doubleSpinBox_electron_voltage.setValue(electron_configuration["voltage"])                                 # V
        self.doubleSpinBox_electron_current.setValue(electron_configuration["current"] * constants.SI_TO_NANO)         # nA
        self.spinBox_electron_resolution_x.setValue(electron_configuration["resolution"][0])
        self.spinBox_electron_resolution_y.setValue(electron_configuration["resolution"][1])
        self.doubleSpinBox_electron_hfw.setValue(electron_configuration["hfw"] * constants.SI_TO_MICRO)                 # um
        self.doubleSpinBox_electron_dwell_time.setValue(electron_configuration["dwell_time"] * constants.SI_TO_MICRO)   # us
        self.lineEdit_electron_detector_mode.setText(electron_configuration["detector_mode"])
        self.lineEdit_electron_detector_type.setText(electron_configuration["detector_type"])

        # ion
        ion_configuration = self.configuration["ion"]
        self.spinBox_ion_column_tilt.setValue(ion_configuration["column_tilt"])                                         # deg
        self.doubleSpinBox_ion_eucentric_height.setValue(ion_configuration["eucentric_height"] * constants.SI_TO_MILLI) # mm
        self.doubleSpinBox_ion_voltage.setValue(ion_configuration["voltage"])                                           # V
        self.doubleSpinBox_ion_current.setValue(ion_configuration["current"] * constants.SI_TO_NANO)               # nA
        self.spinBox_ion_resolution_x.setValue(ion_configuration["resolution"][0])
        self.spinBox_ion_resolution_y.setValue(ion_configuration["resolution"][1])
        self.doubleSpinBox_ion_hfw.setValue(ion_configuration["hfw"] * constants.SI_TO_MICRO)                         # um
        self.doubleSpinBox_ion_dwell_time.setValue(ion_configuration["dwell_time"] * constants.SI_TO_MICRO)           # us
        self.lineEdit_ion_detector_mode.setText(ion_configuration["detector_mode"])
        self.lineEdit_ion_detector_type.setText(ion_configuration["detector_type"])
        
        # imaging
        imaging_configuration =  self.configuration["imaging"]
        self.comboBox_imaging_beam_type.setCurrentText(imaging_configuration["beam_type"])
        self.spinBox_imaging_resolution_x.setValue(imaging_configuration["resolution"][0])
        self.spinBox_imaging_resolution_y.setValue(imaging_configuration["resolution"][1])
        self.doubleSpinBox_imaging_hfw.setValue(imaging_configuration["hfw"] * constants.SI_TO_MICRO)                   # um
        self.doubleSpinBox_imaging_dwell_time.setValue(imaging_configuration["dwell_time"] * constants.SI_TO_MICRO)     # us
        self.doubleSpinBox_imaging_current.setValue(imaging_configuration["imaging_current"] * constants.SI_TO_NANO)    # nA
        self.checkBox_imaging_autocontrast.setChecked(imaging_configuration["autocontrast"])
        self.checkBox_imaging_autogamma.setChecked(imaging_configuration["autogamma"])
        self.checkBox_imaging_autosave.setChecked(imaging_configuration["save"])

        # milling
        milling_configuration = self.configuration["milling"]
        self.doubleSpinBox_milling_voltage.setValue(milling_configuration["milling_voltage"])                                   # V
        self.doubleSpinBox_milling_current.setValue(milling_configuration["milling_current"] * constants.SI_TO_NANO)            # nA
        self.doubleSpinBox_milling_dwell_time.setValue(milling_configuration["dwell_time"] * constants.SI_TO_MICRO)     # us
        self.doubleSpinBox_milling_rate.setValue(milling_configuration["rate"])                                         # um3/s
        self.doubleSpinBox_milling_spot_size.setValue(milling_configuration["spot_size"] * constants.SI_TO_NANO)        # nm
        self.lineEdit_milling_preset.setText(milling_configuration["preset"])

        # subsystems
        subsystem_configuration = self.configuration
        # electron
        self.checkBox_subsystems_electron_enabled.setChecked(subsystem_configuration["electron"]["enabled"])
        # ion
        self.checkBox_subsystems_ion_enabled.setChecked(subsystem_configuration["ion"]["enabled"])
        self.checkBox_subsystems_ion_plasma.setChecked(subsystem_configuration["ion"]["plasma"])
        # stage
        self.checkBox_subsystems_stage_enabled.setChecked(subsystem_configuration["stage"]["enabled"])
        self.checkBox_subsystems_stage_rotation.setChecked(subsystem_configuration["stage"]["rotation"])
        self.checkBox_subsystems_stage_tilt.setChecked(subsystem_configuration["stage"]["tilt"])
        # manipulator
        self.checkBox_subsystems_manipulator_enabled.setChecked(subsystem_configuration["manipulator"]["enabled"])
        self.checkBox_subsystems_manipulator_rotation.setChecked(subsystem_configuration["manipulator"]["rotation"])
        self.checkBox_subsystems_manipulator_tilt.setChecked(subsystem_configuration["manipulator"]["tilt"])
        # gis
        self.checkBox_subsystems_gis_enabled.setChecked(subsystem_configuration["gis"]["enabled"])
        self.checkBox_subsystems_gis_multichem_enabled.setChecked(subsystem_configuration["gis"]["multichem"])
        self.checkBox_subsystems_gis_sputter_coater_enabled.setChecked(subsystem_configuration["gis"]["sputter_coater"])


    def read_configuration_from_ui(self) -> dict:

        # core
        info_configuration = self.configuration["info"]
        info_configuration["name"] = self.lineEdit_configuration_name.text()
        info_configuration["ip_address"] = self.lineEdit_configuration_ip_address.text()
        info_configuration["manufacturer"] = self.comboBox_configuration_manufacturer.currentText()

        # stage
        stage_configuration = self.configuration["stage"]
        stage_configuration["enabled"] = self.checkBox_subsystems_stage_enabled.isChecked()
        stage_configuration["rotation"] = self.checkBox_subsystems_stage_rotation.isChecked()
        stage_configuration["tilt"] = self.checkBox_subsystems_stage_tilt.isChecked()
        stage_configuration["rotation_reference"] = self.spinBox_stage_rotation_reference.value()
        stage_configuration["rotation_180"] = self.spinBox_stage_rotation_180.value()
        stage_configuration["shuttle_pre_tilt"] = self.spinBox_stage_shuttle_pre_tilt.value()
        stage_configuration["manipulator_height_limit"] = self.doubleSpinBox_stage_manipulator_height_limit.value() * constants.MILLI_TO_SI

        # electron
        electron_configuration = self.configuration["electron"]
        electron_configuration["enabled"] = self.checkBox_subsystems_electron_enabled.isChecked()
        electron_configuration["column_tilt"] = self.spinBox_electron_column_tilt.value()
        electron_configuration["eucentric_height"] = self.doubleSpinBox_electron_eucentric_height.value() * constants.MILLI_TO_SI
        electron_configuration["voltage"] = self.doubleSpinBox_electron_voltage.value()
        electron_configuration["current"] = self.doubleSpinBox_electron_current.value() * constants.NANO_TO_SI
        electron_configuration["resolution"] = [self.spinBox_electron_resolution_x.value(), self.spinBox_electron_resolution_y.value()]
        electron_configuration["hfw"] = self.doubleSpinBox_electron_hfw.value() * constants.MICRO_TO_SI
        electron_configuration["dwell_time"] = self.doubleSpinBox_electron_dwell_time.value() * constants.MICRO_TO_SI
        electron_configuration["detector_mode"] = self.lineEdit_electron_detector_mode.text()
        electron_configuration["detector_type"] = self.lineEdit_electron_detector_type.text()

        # ion
        ion_configuration = self.configuration["ion"]
        ion_configuration["enabled"] = self.checkBox_subsystems_ion_enabled.isChecked()
        ion_configuration["plasma"] = self.checkBox_subsystems_ion_plasma.isChecked()
        ion_configuration["column_tilt"] = self.spinBox_ion_column_tilt.value()
        ion_configuration["eucentric_height"] = self.doubleSpinBox_ion_eucentric_height.value() * constants.MILLI_TO_SI
        ion_configuration["voltage"] = self.doubleSpinBox_ion_voltage.value()
        ion_configuration["current"] = self.doubleSpinBox_ion_current.value() * constants.NANO_TO_SI
        ion_configuration["resolution"] = [self.spinBox_ion_resolution_x.value(), self.spinBox_ion_resolution_y.value()]
        ion_configuration["hfw"] = self.doubleSpinBox_ion_hfw.value() * constants.MICRO_TO_SI
        ion_configuration["dwell_time"] = self.doubleSpinBox_ion_dwell_time.value() * constants.MICRO_TO_SI
        ion_configuration["detector_mode"] = self.lineEdit_ion_detector_mode.text()
        ion_configuration["detector_type"] = self.lineEdit_ion_detector_type.text()

        # manipulator
        manipulator_configuration = self.configuration["manipulator"]
        manipulator_configuration["enabled"] = self.checkBox_subsystems_manipulator_enabled.isChecked()
        manipulator_configuration["rotation"] = self.checkBox_subsystems_manipulator_rotation.isChecked()
        manipulator_configuration["tilt"] = self.checkBox_subsystems_manipulator_tilt.isChecked()
        
        # gis
        gis_configuration = self.configuration["gis"]
        gis_configuration["enabled"] = self.checkBox_subsystems_gis_enabled.isChecked()
        gis_configuration["multichem"] = self.checkBox_subsystems_gis_multichem_enabled.isChecked()
        gis_configuration["sputter_coater"] = self.checkBox_subsystems_gis_sputter_coater_enabled.isChecked()

        # imaging
        imaging_configuration =  self.configuration["imaging"]
        imaging_configuration["beam_type"] = self.comboBox_imaging_beam_type.currentText()
        imaging_configuration["resolution"] = [self.spinBox_imaging_resolution_x.value(), self.spinBox_imaging_resolution_y.value()]
        imaging_configuration["hfw"] = self.doubleSpinBox_imaging_hfw.value() * constants.MICRO_TO_SI
        imaging_configuration["dwell_time"] = self.doubleSpinBox_imaging_dwell_time.value() * constants.MICRO_TO_SI
        imaging_configuration["imaging_current"] = self.doubleSpinBox_imaging_current.value() * constants.NANO_TO_SI
        imaging_configuration["autocontrast"] = self.checkBox_imaging_autocontrast.isChecked()
        imaging_configuration["autogamma"] = self.checkBox_imaging_autogamma.isChecked()
        imaging_configuration["save"] = self.checkBox_imaging_autosave.isChecked()

        # milling
        milling_configuration = self.configuration["milling"]
        milling_configuration["milling_voltage"] = self.doubleSpinBox_milling_voltage.value()
        milling_configuration["milling_current"] = self.doubleSpinBox_milling_current.value() * constants.NANO_TO_SI
        milling_configuration["dwell_time"] = self.doubleSpinBox_milling_dwell_time.value() * constants.MICRO_TO_SI
        milling_configuration["rate"] = self.doubleSpinBox_milling_rate.value()
        milling_configuration["spot_size"] = self.doubleSpinBox_milling_spot_size.value() * constants.NANO_TO_SI
        milling_configuration["preset"] = self.lineEdit_milling_preset.text()

        # configuration
        configuration = {
            "info": info_configuration,
            "stage": stage_configuration,
            "electron": electron_configuration,
            "ion": ion_configuration,
            "manipulator": manipulator_configuration,
            "gis": gis_configuration,         
            "imaging": imaging_configuration,
            "milling": milling_configuration,
        }

        return configuration
    



def main():
    path = cfg.DEFAULT_CONFIGURATION_PATH
    # widget viewer
    viewer = napari.Viewer(ndisplay=2)
    microscope_configuration = FibsemMicroscopeConfigurationWidgetBase(path=path, viewer=viewer)
    viewer.window.add_dock_widget(
        microscope_configuration, 
        area="right", 
        add_vertical_stretch=False,
        name=f"OpenFIBSEM v{fibsem.__version__} Microscope Configuration",
    )
    napari.run()


if __name__ == "__main__":
    main()


import logging
import os
import sys
import glob

# import fibsem
import napari
import napari.utils.notifications
from PIL import Image
from PyQt5 import QtWidgets
import dask.array as da
from fibsem.segmentation import config as cfg
from fibsem.segmentation import utils as seg_utils
from fibsem.segmentation import train as seg_train

from fibsem.ui.qtdesigner_files import FibsemModelTrainingWidget
import yaml
from fibsem.segmentation.config import CLASS_COLORS


# BASE_PATH = os.path.join(os.path.dirname(fibsem.__file__), "config")
from PyQt5.QtCore import pyqtSignal

from napari.qt.threading import thread_worker


class FibsemModelTrainingWidget(FibsemModelTrainingWidget.Ui_Form, QtWidgets.QWidget):
    ui_signal = pyqtSignal(dict)

    def __init__(
        self,
        viewer: napari.Viewer,
        parent=None,
    ):
        super(FibsemModelTrainingWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.viewer = viewer

        self.setup_connections()

    def setup_connections(self):
        self.pushButton_train_model.clicked.connect(self.train_model)

        self.comboBox_encoder.addItems(seg_utils.unet_encoders)

        self.ui_signal.connect(self.train_model_yielded)

        # NOTE: Setup your config.yml file
        with open(cfg.CONFIG_PATH, "r") as f:
            config = yaml.safe_load(f)

        checkpoint = config["checkpoint"] if config["checkpoint"] is not None else ""

        self.lineEdit_data_path.setText(config["data_paths"][0])
        self.lineEdit_label_path.setText(config["label_paths"][0])
        self.lineEdit_save_path.setText(config["save_path"])
        self.comboBox_encoder.setCurrentText(config["encoder"])
        self.lineEdit_checkpoint.setText(checkpoint)
        self.spinBox_num_classes.setValue(config["num_classes"])
        self.spinBox_batch_size.setValue(config["batch_size"])
        self.spinBox_epochs.setValue(config["epochs"])
        self.doubleSpinBox_lr.setValue(config["lr"])
        self.checkBox_use_wandb.setChecked(config["wandb"])
        self.lineEdit_wandb_project.setText(config["wandb_project"])
        self.lineEdit_wandb_entity.setText(config["wandb_entity"])

    def train_model(self):
        checkpoint = self.lineEdit_checkpoint.text()
        checkpoint = None if checkpoint == "" else checkpoint

        config = {
            "data_paths": [self.lineEdit_data_path.text()],
            "label_paths": [self.lineEdit_label_path.text()],
            "save_path": self.lineEdit_save_path.text(),
            "encoder": self.comboBox_encoder.currentText(),
            "checkpoint": checkpoint,
            "num_classes": int(self.spinBox_num_classes.value()),
            "batch_size": self.spinBox_batch_size.value(),
            "epochs": self.spinBox_epochs.value(),
            "lr": self.doubleSpinBox_lr.value(),
            "wandb": self.checkBox_use_wandb.isChecked(),
            "wandb_project": self.lineEdit_wandb_project.text(),
            "wandb_entity": self.lineEdit_wandb_entity.text(),
        }

        self.label_info.setText("Model Training...")
        self.pushButton_train_model.setEnabled(False)
        worker = self.train_worker(config)
        worker.finished.connect(self.train_model_finished)
        worker.start()

    @thread_worker
    def train_worker(self, config: dict):
        ################################## VALIDATE CONFIG ##################################
        seg_utils.validate_config(config)

        ################################## SETUP W&B ##################################
        seg_train._setup_wandb(config)

        ################################## LOAD DATASET ##################################
        train_data_loader, val_data_loader = seg_train._setup_dataset(config)

        ################################## LOAD MODEL ##################################

        model, optimizer, device = seg_train._setup_model(config)
        
        self.viewer.layers.clear()
        # self.viewer.grid.enabled = True

        seg_train.train_model(
            model,
            device,
            optimizer,
            train_data_loader,
            val_data_loader,
            config=config,
        )

    def train_model_yielded(self, info: dict):

        if info["stage"] == "end":
            self.epoch = info["epoch"]
            self.epochs = info["epochs"]
            self.train_loss = info["train_loss"]
            self.val_loss = info["val_loss"]
        if info["stage"] == "train":
            self.train_loss = info["train_loss"]
        if info["stage"] == "val":
            self.val_loss = info["val_loss"]
        msg = f"Epoch: {self.epoch}/{self.epochs} | Train Loss: {self.train_loss:.4f} | Val Loss: {self.val_loss:.4f}"
        self.label_info.setText(msg)

        if info["stage"] in ["train", "val"]:
            
            try:
                self._image_layer.data = info["image"]
            except:
                self._image_layer = self.viewer.add_image(
                    info["image"],
                    name="Image",
                    opacity=0.7,
                    blending="additive",
                )
            try:
                self._label_layer.data = info["gt"]
            except:
                self._label_layer = self.viewer.add_labels(
                    info["gt"],
                    name="Mask",    
                    opacity=0.7,
                    blending="additive",
                    color=CLASS_COLORS,
                )
            try:
                self._pred_layer.data = info["pred"]
            except:
                self._pred_layer = self.viewer.add_labels(
                    info["pred"],
                    name="Prediction",
                    opacity=0.7,
                    blending="additive",
                    color=CLASS_COLORS,
                )

    def train_model_finished(self):
        msg = "Train Model Finished!"
        napari.utils.notifications.show_info(msg)
        self.label_info.setText(msg)
        self.pushButton_train_model.setEnabled(True)

    def label_info(self, text: str):
        self.label_info.setText(text)


def main():
    viewer = napari.Viewer()
    widget = FibsemModelTrainingWidget(viewer=viewer)
    viewer.window.add_dock_widget(widget, area="right")
    napari.run()


if __name__ == "__main__":
    main()


# TODO: stop training button



import logging
import os

import napari
import napari.utils.notifications
import numpy as np
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QMessageBox

from fibsem import constants, conversions
from fibsem.microscope import (DemoMicroscope, FibsemMicroscope,
                               TescanMicroscope, ThermoMicroscope)
from fibsem.structures import (BeamType, FibsemManipulatorPosition,
                               MicroscopeSettings)
from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
from fibsem.ui.qtdesigner_files import FibsemManipulatorWidget
from fibsem.ui.utils import message_box_ui
from fibsem.ui import _stylesheets
from fibsem import config as cfg





class FibsemManipulatorWidget(FibsemManipulatorWidget.Ui_Form, QtWidgets.QWidget):
    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        viewer: napari.Viewer = None,
        image_widget: FibsemImageSettingsWidget = None, 
        parent=None,
    ):
        super(FibsemManipulatorWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.microscope = microscope
        self.settings = settings
        self.viewer = viewer
        self.image_widget = image_widget
        self.saved_positions = {}

        self.setup_connections()

        self.update_ui()

        _THERMO = isinstance(self.microscope, (ThermoMicroscope,DemoMicroscope))
        _TESCAN = isinstance(self.microscope, (TescanMicroscope))


        if _THERMO:
            
            try:
                self.microscope._get_saved_manipulator_position("PARK")
                self.microscope._get_saved_manipulator_position("EUCENTRIC")
                self.savedPosition_combobox.addItems(["PARK", "EUCENTRIC"])
             
            except :
                napari.utils.notifications.show_warning("Error loading PARK and EUCENTRIC positions, calibration of manipulator is possibly needed.")
                            
            self.move_type_comboBox.currentIndexChanged.connect(self.change_move_type)
            self.move_type_comboBox.setCurrentIndex(0)
            self.change_move_type()
            self.dR_spinbox.setEnabled(False)
            self.dR_spinbox.hide()
            self.dr_label.hide()
            self.calibrated_status_label.hide()

        if _TESCAN:

            self.tescan_calibration = cfg.load_tescan_manipulator_calibration()

            self._initialise_calibration()
            self.move_type_comboBox.hide()
            self.beam_type_label.hide()
            self.beam_type_combobox.hide()
            # self.insertManipulator_button.hide()
            # self.manipulatorStatus_label.hide()
            self.savedPosition_combobox.addItem("Standby")
            self.savedPosition_combobox.addItem("Working")

        manipulator_inserted = self.microscope.get_manipulator_state()
        self._hide_show_buttons(manipulator_inserted)
        self.insertManipulator_button.setText("Retract" if manipulator_inserted else "Insert")
        self.manipulatorStatus_label.setText("Manipulator Status: Inserted" if manipulator_inserted else "Manipulator Status: Retracted")


    def _initialise_calibration(self):

        is_calibrated = self.tescan_calibration["calibrated"]

        if not is_calibrated:

            self.calibrated_status_label.setText("Not Calibrated, Please run the calibration tool from the tool menu")
            self.insertManipulator_button.setEnabled(False)
            self._hide_show_buttons(show=False)

    def _check_manipulator_positions_setup(self):

        is_calibrated = self.tescan_calibration["calibrated"]

        if is_calibrated:
            response = message_box_ui(title="Manipulator Positions Already Calibrated", text="Manipulator Positions are already calibrated, would you like to recalibrate?")
            response = not response
        else:
            response = False

        return response


    def calibrate_manipulator_positions(self):

        if not isinstance(self.microscope,TescanMicroscope):
            message_box_ui(title="Not Available", text="Manipulator Position Calibration is only available for Tescan Microscopes", buttons=QMessageBox.Ok)
            return

        response = self._check_manipulator_positions_setup()

        if not response:
            
            ok_to_cal = message_box_ui(title="Manipulator Position calibration", 
                                       text="This tool calibrates the positions of the manipulator, it will switch between the parking, standby and working positions rapidly, please ensure it is safe to do so. If not please click no, otherwise press yes to continue")
                                      
            if ok_to_cal:

                calibration = self.tescan_calibration

                for position in ["parking", "standby", "working"]:
                    logging.info(f"Calibrating Manipulator {position} position")
                    self.microscope.insert_manipulator(position)
                    manipulator_loc = self.microscope.get_manipulator_position()
                    calibration[position]["x"] = manipulator_loc.x
                    calibration[position]["y"] = manipulator_loc.y
                    calibration[position]["z"] = manipulator_loc.z
                
                calibration["calibrated"] = True
                cfg.save_tescan_manipulator_calibration(calibration)
                self.tescan_calibration = cfg.load_tescan_manipulator_calibration()

                message_box_ui(title="Manipulator Position calibration", 
                               text="Manipulator Positions calibrated successfully", 
                               buttons=QMessageBox.Ok)

                self.update_ui_state()
                


    def change_move_type(self):

        if self.move_type_comboBox.currentText() == "Relative Move":
            self.dZ_spinbox.setEnabled(True)
            self.dZ_spinbox.show()
            self.dz_label.show()
            self.beam_type_label.hide()
            self.beam_type_combobox.hide()
        else:
            self.dZ_spinbox.setEnabled(False)
            self.dZ_spinbox.hide()
            self.dz_label.hide()
            self.beam_type_label.show()
            self.beam_type_combobox.show()

    def update_ui_state(self):

        if isinstance(self.microscope, (ThermoMicroscope,DemoMicroscope)):
            is_calibrated = True

        if isinstance(self.microscope, (TescanMicroscope)):
            is_calibrated = self.tescan_calibration["calibrated"]
        
        is_inserted = self.microscope.get_manipulator_state()
        self.insertManipulator_button.setEnabled(is_calibrated)
        self.moveRelative_button.setEnabled(is_calibrated)
        self.addSavedPosition_button.setEnabled(is_calibrated)
        self.goToPosition_button.setEnabled(is_calibrated)
        self._hide_show_buttons(show=is_inserted)
        self.manipulatorStatus_label.setText("Manipulator Status: Inserted" if is_inserted else "Manipulator Status: Retracted")
        self.insertManipulator_button.setText("Insert" if not is_inserted else "Retract")
        self.calibrated_status_label.setText("Calibrated") 

    def update_ui(self):

        self.dX_spinbox.setValue(0)
        self.dY_spinbox.setValue(0)
        self.dZ_spinbox.setValue(0)
        self.dR_spinbox.setValue(0)


        
    def setup_connections(self):

        self.insertManipulator_button.clicked.connect(self.insert_retract_manipulator)
        self.insertManipulator_button.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.addSavedPosition_button.clicked.connect(self.add_saved_position)
        self.addSavedPosition_button.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.goToPosition_button.clicked.connect(self.move_to_saved_position)
        self.goToPosition_button.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.moveRelative_button.clicked.connect(self.move_relative)
        self.moveRelative_button.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)

        self.pushButton_refresh_data.clicked.connect(self.refresh_data)
        self.pushButton_refresh_data.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)


    def refresh_data(self):
        self.manipulator_inserted = self.microscope.get_manipulator_state()
        
        self._hide_show_buttons(self.manipulator_inserted)
        self.insertManipulator_button.setText("Retract" if self.manipulator_inserted else "Insert")
        self.insertManipulator_button.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE if self.manipulator_inserted else _stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.manipulatorStatus_label.setText("Manipulator Status: Inserted" if self.manipulator_inserted else "Manipulator Status: Retracted")        
        

    def move_relative(self):
        """Move the manipulator relative to its current position."""

        dx = self.dX_spinbox.value() * constants.MICRO_TO_SI
        dy = self.dY_spinbox.value() * constants.MICRO_TO_SI
        dz = self.dZ_spinbox.value() * constants.MICRO_TO_SI
        dr = self.dR_spinbox.value() * constants.DEGREES_TO_RADIANS
        beam_type = getattr(BeamType,self.beam_type_combobox.currentText())

        if self.move_type_comboBox.currentText() == "Relative Move" or isinstance(self.microscope, (TescanMicroscope)):

            try:
                position = FibsemManipulatorPosition(x=dx, y=dy, z=dz, r=dr, coordinate_system="STAGE")  # TODO migrate to raw manipulator movements
                self.microscope.move_manipulator_relative(position=position)
            except Exception as e:
                error_message = f"Error moving manipulator (Relative): {str(e)}"
                logging.error(error_message)
                napari.utils.notifications.show_error(error_message)
            
        else:
            try:
                self.microscope.move_manipulator_corrected(dx=dx,dy=dy,beam_type=beam_type)
            except Exception as e:
                error_message = f"Error moving manipulator (Corrected): {str(e)}"
                logging.error(error_message)
                napari.utils.notifications.show_error(error_message)

        self.update_ui()      
  
    def _hide_show_buttons(self,show:bool = True):

        # show = False

        self.move_type_comboBox.setVisible(show)
        self.dX_spinbox.setVisible(show)
        self.dY_spinbox.setVisible(show)
        self.dZ_spinbox.setVisible(show)
        self.dR_spinbox.setVisible(show)
        self.dz_label.setVisible(show)
        self.dx_label.setVisible(show)
        self.dy_label.setVisible(show)
        self.dr_label.setVisible(show)
        self.beam_type_combobox.setVisible(show)
        self.beam_type_label.setVisible(show)
        self.moveRelative_button.setVisible(show)
        self.addSavedPosition_button.setVisible(show)
        self.goToPosition_button.setVisible(show)
        self.savedPositionName_lineEdit.setVisible(show)
        self.savedPosition_combobox.setVisible(show)




    def insert_retract_manipulator(self):
        
        if self.microscope.get_manipulator_state():

            self.microscope.retract_manipulator()
            self.insertManipulator_button.setText("Insert")
            self.manipulatorStatus_label.setText("Manipulator Status: Retracted")
            self.insertManipulator_button.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.update_ui()
            self._hide_show_buttons(show=False)
        
        else:

            self.microscope.insert_manipulator()
            self.insertManipulator_button.setText("Retract")
            self.manipulatorStatus_label.setText("Manipulator Status: Inserted")
            self.insertManipulator_button.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE)
            self.update_ui()
            self._hide_show_buttons(show=True)



    def add_saved_position(self):

        if self.savedPositionName_lineEdit.text() == "":
            _ = message_box_ui(
                title="No name.",
                text="Please enter a position name.",
                buttons=QMessageBox.Ok,
            )
            return
        name = self.savedPositionName_lineEdit.text()
        position = self.microscope.get_manipulator_position()
        self.saved_positions[name] = position
        logging.info(f"Saved position {name} at {position}")
        self.savedPosition_combobox.addItem(name)
        self.savedPositionName_lineEdit.clear()

    def move_to_saved_position(self):
        name = self.savedPosition_combobox.currentText()

        if name in ["Parking","Standby","Working"] and isinstance(self.microscope, (TescanMicroscope)):
            self.microscope.insert_manipulator(name=name)
            position = self.microscope.get_manipulator_position()
            logging.info(f"Moved to saved position {name} at {position}")
            self.update_ui()
            return
        elif name in ["PARK", "EUCENTRIC"] and isinstance(self.microscope, (ThermoMicroscope, DemoMicroscope)):
            position = self.microscope._get_saved_manipulator_position(name)
        else: 
            position = self.saved_positions[name]

        logging.info(f"Moving to saved position {name} at {position}")
        self.microscope.move_manipulator_absolute(position=position)
        self.update_ui()



def main():

    viewer = napari.Viewer(ndisplay=2)
    manipulator_widget = FibsemManipulatorWidget()
    viewer.window.add_dock_widget(
        manipulator_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'detection_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(1016, 692)
        self.frame_main = QtWidgets.QFrame(Dialog)
        self.frame_main.setGeometry(QtCore.QRect(0, 10, 994, 658))
        self.frame_main.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_main.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_main.setObjectName("frame_main")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.frame_main)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.frame_image = QtWidgets.QFrame(self.frame_main)
        self.frame_image.setMaximumSize(QtCore.QSize(1536, 1024))
        self.frame_image.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_image.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_image.setObjectName("frame_image")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.frame_image)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_image = QtWidgets.QLabel(self.frame_image)
        self.label_image.setText("")
        self.label_image.setObjectName("label_image")
        self.gridLayout_3.addWidget(self.label_image, 0, 0, 1, 1)
        self.gridLayout_2.addWidget(self.frame_image, 1, 1, 2, 1)
        self.frame_parameters = QtWidgets.QFrame(self.frame_main)
        self.frame_parameters.setMaximumSize(QtCore.QSize(200, 16777215))
        self.frame_parameters.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_parameters.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_parameters.setObjectName("frame_parameters")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.frame_parameters)
        self.gridLayout_4.setVerticalSpacing(6)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_header_2 = QtWidgets.QLabel(self.frame_parameters)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_header_2.setFont(font)
        self.label_header_2.setObjectName("label_header_2")
        self.gridLayout_4.addWidget(self.label_header_2, 1, 1, 1, 1)
        self.comboBox_detection_type = QtWidgets.QComboBox(self.frame_parameters)
        self.comboBox_detection_type.setObjectName("comboBox_detection_type")
        self.gridLayout_4.addWidget(self.comboBox_detection_type, 2, 1, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_4.addItem(spacerItem, 15, 1, 1, 2)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.gridLayout_4.addItem(spacerItem1, 19, 1, 1, 2)
        self.label_movement = QtWidgets.QLabel(self.frame_parameters)
        self.label_movement.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_movement.setObjectName("label_movement")
        self.gridLayout_4.addWidget(self.label_movement, 17, 1, 1, 2)
        self.pushButton_continue = QtWidgets.QPushButton(self.frame_parameters)
        self.pushButton_continue.setMinimumSize(QtCore.QSize(0, 40))
        self.pushButton_continue.setObjectName("pushButton_continue")
        self.gridLayout_4.addWidget(self.pushButton_continue, 20, 1, 1, 2)
        self.label_movement_header = QtWidgets.QLabel(self.frame_parameters)
        self.label_movement_header.setTextFormat(QtCore.Qt.AutoText)
        self.label_movement_header.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label_movement_header.setObjectName("label_movement_header")
        self.gridLayout_4.addWidget(self.label_movement_header, 16, 1, 1, 1)
        self.gridLayout_2.addWidget(self.frame_parameters, 1, 0, 2, 1)
        self.label_title = QtWidgets.QLabel(self.frame_main)
        self.label_title.setMaximumSize(QtCore.QSize(16777215, 50))
        font = QtGui.QFont()
        font.setPointSize(26)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setTextFormat(QtCore.Qt.PlainText)
        self.label_title.setObjectName("label_title")
        self.gridLayout_2.addWidget(self.label_title, 0, 0, 1, 2)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label_header_2.setText(_translate("Dialog", "Detection Type"))
        self.label_movement.setText(_translate("Dialog", "TextLabel"))
        self.pushButton_continue.setText(_translate("Dialog", "Continue"))
        self.label_movement_header.setText(_translate("Dialog", "TextLabel"))
        self.label_title.setText(_translate("Dialog", "AutoLiftout Detection"))

